/* $Id: ftp.c,v 1.3 2000/02/29 04:15:47 leonard Exp $ */
/* Copyright David Leonard, 2000. All rights reserved. */
#include <stdio.h>
#include <unistd.h>
#include <stdarg.h>
#include <string.h>
#include <err.h>
#include <errno.h>
#include <stdlib.h>
#include <ctype.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netdb.h>
#include "ftp.h"
#include "finderror.h"

/* Refer to RFC 959 for details on the FTP protocol. */

extern int debugflag;

const char CRLF[2] = { '\r', '\n' };

/*
 * Open a connection to an FTP server, and login as the given user
 * with the given password
 */
struct ftp_ctx *
ftp_new_ctx(hostname, username, password, passive)
	const char *hostname, *username, *password;
	int passive;
{
	struct ftp_ctx *ctx;

	ctx = (struct ftp_ctx *)malloc(sizeof (struct ftp_ctx));
	if (ctx == NULL) {
		warn("malloc");
		return NULL;
	}
	strncpy(ctx->username, username, sizeof ctx->username);
	strncpy(ctx->hostname, hostname, sizeof ctx->hostname);
	strncpy(ctx->password, password, sizeof ctx->password);
	ctx->passive = passive;
	ctx->f = NULL;

	return ctx;
}

int
ftp_connect(ctx)
	struct ftp_ctx *ctx;
{
	int fd;
	int error;
	struct addrinfo hint, *res, *ai;
	struct ftp_response resp;

	/* Close existing connection */
	if (ctx->f) {
		fclose(ctx->f);
		ctx->f = NULL;
	}

	if (debugflag)
		fprintf(stderr, "FTP: connecting to %s\n", ctx->hostname);

	/* Convert the hostname into a socket address */
	bzero(&hint, sizeof hint);
	hint.ai_family = PF_UNSPEC;	/* PF_INET */
	hint.ai_socktype = SOCK_STREAM;
	hint.ai_protocol = 0;

	error = getaddrinfo(ctx->hostname, "ftp", &hint, &res);
	if (error) {
		warnx("%s:ftp: %s", ctx->hostname, gai_strerror(error));
		errno = ENOENT;
		return -1;
	}

	/* Try connecting to the address */
	fd = -1;
	for (ai = res; ai; ai = ai->ai_next) {
		fd = socket(ai->ai_family, ai->ai_socktype, ai->ai_protocol);
		if (fd < 0) 
			continue;
		if (connect(fd, ai->ai_addr, (socklen_t)ai->ai_addrlen) == 0)
			break;
		close(fd);
		fd = -1;
	}
	freeaddrinfo(res);
	if (fd < 0)
		return -1;

	/* The FTP login protocol */
	ctx->f = fdopen(fd, "a+");

	/* Receive the welcome message */
	if (ftp_getresp(ctx, &resp) == -1 || resp.code / 100 != 2)
		goto login_failed;
	/* Send username */
	if (ftp_sendcmd(ctx, &resp, "USER %s", ctx->username) == -1 && 
		resp.code != 331)
		goto login_failed;
	/* Send password */
	if (ftp_sendcmd(ctx, &resp, "PASS %s", ctx->password) == -1 ||
		resp.code / 100 != 2)
		goto login_failed;

	return 0;

login_failed:
	fclose(ctx->f);
	ctx->f = NULL;
	warnx("%s@%s: %s", ctx->username, ctx->hostname, resp.msg);
	errno = EACCES;
	return -1;
}

/* 
 * Get a response string into a ftp_response structure.
 * Also handles multiple lines
 */
static int
getmultiline(ctx, resp)
	struct ftp_ctx *ctx;
	struct ftp_response *resp;
{
	size_t left;
	size_t len;
	char *line;
	char *p;

	if (!ctx->f) {
		errno = ENOTCONN;
		return -1;
	}

	left = sizeof resp->msg - 1; /* leave space for nul terminator */
	p = resp->msg;
	resp->code = 599;

	line = fgetln(ctx->f, &len);
	if (line == NULL)
		goto end;
	if (len > 2 && line[len-1] == '\n' && line[len-2] == '\r')
		len -= 2;
	if (debugflag)
		fprintf(stderr, "-> %.*s\n", (int)len, line);
	if (len < 4 || (line[3] != '-' && line[3] != ' ') ||
	    !isdigit(line[0]) || !isdigit(line[1]) || !isdigit(line[2]))
		goto end;
	resp->code = atoi(line);

	/* Copy primary response into response buffer */
	if (len - 4 > left)
		len = left + 4;
	memcpy(p, line + 4, len - 4);
	p += (len - 4);
	left -= (len - 4);

	if (line[3] == '-')
		while((line = fgetln(ctx->f, &len)) != NULL) {
			if (len > 2 && line[len-1] == '\n' && 
			    line[len-2] == '\r')
				len -= 2;
			if (debugflag)
				fprintf(stderr, "-> %.*s\n", (int)len, line);
			if (left) {
				*p++ = '\n';
				left--;
			}
			if (len > left)
				len = left;
			memcpy(p, line, len);
			p += len;
			left -= len;

			/* Response ends when we see the response code again */
			if (len > 3 && line[3] == ' ' && atoi(line) == 
			    resp->code)
				break;
		}

    end:
	if (feof(ctx->f)) {
		fclose(ctx->f);
		ctx->f = NULL;
		errno = ENOTCONN;
		return -1;
	}

	*p = '\0';
	if (debugflag)
		fprintf(stderr, "-> [%03d]\n", resp->code);
	return 0;
}

/* Get a simple response code, returning 0 if its not an error */
int
ftp_getresp(ctx, resp)
	struct ftp_ctx *ctx;
	struct ftp_response *resp;
{
	int ret;

	ret = getmultiline(ctx, resp);
	if (ret == 0 && resp->code >= 300)
		return finderror(resp->msg, EIO);
	return ret;
}

/* Get a simple response code, return 0 if it is successful */
int
ftp_voidresp(ctx, resp)
	struct ftp_ctx *ctx;
	struct ftp_response *resp;
{
	int ret;

	ret = ftp_getresp(ctx, resp);
	if (resp->code / 100 != 2)
		return resp->code;
	return ret;
}

/* Abort a transfer, returning 0 if abort was successful */
int
ftp_abort(ctx)
	struct ftp_ctx *ctx;
{
	static char abort[] = "ABORT\r\n";
	struct ftp_response resp;
	int ret;

	send(fileno(ctx->f), abort, strlen(abort), MSG_OOB);
	ret = getmultiline(ctx, &resp);
	if (ret == 0 && (resp.code != 426 && resp.code != 226))
		return resp.code;
	return ret;
}

/* Send a single command, and get the response. Returns 0 if errcode < 400 */
int
ftp_sendcmd(ctx, resp, cmd)
	struct ftp_ctx *ctx;
	struct ftp_response *resp;
	const char *cmd;
{
	int ret;
	va_list ap;

	va_start(ap, cmd);
	ret = ftp_vsendcmd(ctx, resp, cmd, ap);
	va_end(ap);
	return ret;
}


/* Send a single command, and get the response. Returns 0 if errcode < 400 */
int
ftp_vsendcmd(ctx, resp, cmd, ap)
	struct ftp_ctx *ctx;
	struct ftp_response *resp;
	const char *cmd;
	va_list ap;
{

	if (!ctx->f) {
		errno = ENOTCONN;
		return -1;
	}

	if (debugflag) {
		fprintf(stderr, "<- ");
		if (memcmp(cmd, "PASS ", 5) == 0)	/* Hide passwords */
			fprintf(stderr, "PASS *");
		else
			vfprintf(stderr, cmd, ap);
		fprintf(stderr, "\n");
	}
	vfprintf(ctx->f, cmd, ap);
	fwrite(CRLF, sizeof CRLF, 1, ctx->f);
	fflush(ctx->f);
	return ftp_getresp(ctx, resp);
}

/* 
 * Return a file descriptor, set up for a transfer command.
 * or return -1 if there was an error.
 */
int
ftp_ntransfercmd(ctx, resp, cmd)
	struct ftp_ctx *ctx;
	struct ftp_response *resp;
	const char *cmd;
{
	int error;
	int fd = -1;
	socklen_t len;
	struct sockaddr addr;
	struct sockaddr_in *ina = (struct sockaddr_in *)&addr;
	va_list ap;

	if (!ctx->f)
		return -1;

	if (ctx->passive) {
		/*
		 * Ask the server for a passive port to connect to.
		 * Get it back in a 227 response and connect to it.
		 */
		if ((error = ftp_sendcmd(ctx, resp, "PASV")))
			return -1;
		if ((error = ftp_parse227(resp, &addr)))
			return -1;
		if ((fd = socket(addr.sa_family, SOCK_STREAM, 0)) < 0)
			return -1;
		if ((error = connect(fd, &addr, addr.sa_len))) {
			close(fd);
			return -1;
		}
		va_start(ap, cmd);
		error = ftp_vsendcmd(ctx, resp, cmd, ap);
		va_end(ap);
		if (error) {
			close(fd);
			return -1;
		}
	} else {
		/*
		 * Bind to a local port, send the command followed
		 * by the address and then accept on that port
		 */

		len = sizeof addr;
		if (getsockname(fileno(ctx->f), &addr, &len) == -1)
			return -1;

		if (addr.sa_family != AF_INET)
			return -1;

		ina->sin_port = 0;

		fd = socket(PF_INET, SOCK_STREAM, 0);
		if (fd < 0)
			return -1;

		if (bind(fd, &addr, len) == -1)
			goto err;

		if (listen(fd, 1) == -1) 
			goto err;

		len = sizeof addr;
		if (getsockname(fd, &addr, &len) == -1)
			goto err;

		/* Send the PORT command */
		if ((error = ftp_sendcmd(ctx, resp, "PORT %u,%u,%u,%u,%u,%u",
			((unsigned char *)&ina->sin_addr)[0],
			((unsigned char *)&ina->sin_addr)[1],
			((unsigned char *)&ina->sin_addr)[2],
			((unsigned char *)&ina->sin_addr)[3],
			((unsigned char *)&ina->sin_port)[0],
			((unsigned char *)&ina->sin_port)[1]
		)))
			goto err;

		va_start(ap, cmd);
		error = ftp_vsendcmd(ctx, resp, cmd, ap);
		va_end(ap);
		if (error)
			goto err;

		/* Wait for the connection */
		len = sizeof addr;
		if ((error = accept(fd, &addr, &len)) == -1) 
			goto err;
		close(fd);
		fd = error;
	}
	return fd;
  err:
	close(fd);
	return -1;
}

/*
 * Extract the file size from a 150 response. The server doesn't have to
 * supply this information, but some servers do, and it could be handy.
 */
off_t
ftp_parse150(resp)
	struct ftp_response *resp;
{
	char *s;

	if (resp->code != 150)
		return -1;
	/* find the word ' bytes)' then back off to the opening parenthesis */
	s = strstr(resp->msg, " bytes)");
	if (s == NULL)
		return -1;
	while (s != resp->msg && *s != '(')
		s--;
	s++;
	return strtouq(s, NULL, 10);
}

/* Extract a socket address from a 227 response code (eg PASV request) */
int
ftp_parse227(resp, addr)
	struct ftp_response *resp;
	struct sockaddr *addr;
{
	char *s, *n;
	unsigned char bytes[6];
	struct sockaddr_in *ina = (struct sockaddr_in *)addr;
	int i;

	s = strchr(resp->msg, '(');
	if (s == NULL)
		return -1;
	s++;

	for (i = 0; i < 6; i++) {
		if (i && *s++ != ',')
			return -1;
		bytes[i] = strtoul(n = s, &s, 10);
		if (n == s)
			return -1;
	}
	if (*s++ != ')')
		return -1;

	addr->sa_family = AF_INET;
	addr->sa_len = sizeof *ina;
	((unsigned char *)&ina->sin_addr)[0] = bytes[0];
	((unsigned char *)&ina->sin_addr)[1] = bytes[1];
	((unsigned char *)&ina->sin_addr)[2] = bytes[2];
	((unsigned char *)&ina->sin_addr)[3] = bytes[3];
	((unsigned char *)&ina->sin_port)[0] = bytes[4];
	((unsigned char *)&ina->sin_port)[1] = bytes[5];

	return 0;
}

/* Parse the result of a PWD request */
char *
ftp_parse257(resp)
	struct ftp_response *resp;
{
	char *r;

	if (resp->code != 257)
		return NULL;
	if (resp->msg[0] != '"')
		return NULL;
	r = strchr(resp->msg + 1, '"');
	if (r != NULL)
		*r = '\0';
	return resp->msg + 1;
}

/* Request disconnection. (Disconnects regardless of server response.) */
int
ftp_quit(ctx)
	struct ftp_ctx *ctx;
{
	struct ftp_response resp;
	int ret = 0;

	if (ctx->f) {
		ret = ftp_sendcmd(ctx, &resp, "QUIT");
		if (ctx->f) {
			fclose(ctx->f);
			ctx->f = NULL;
		}
	}
	return ret;
}

/* Close connection and deallocate */
void
ftp_free_ctx(ctx)
	struct ftp_ctx *ctx;
{
	if (ctx->f)
		fclose(ctx->f);
	free(ctx);
}
