/* $Id: ftpdir.c,v 1.6 2000/02/29 10:22:18 leonard Exp $ */
/* Copyright David Leonard, 2000. All rights reserved. */
#include <stdio.h>
#include <stdarg.h>
#include <stdlib.h>
#include <fcntl.h>
#include <errno.h>
#include <ctype.h>
#include <unistd.h>
#include <string.h>
#include <err.h>
#include <sys/stat.h>
#include <sys/dirent.h>
#include <sys/socket.h>
#include <xfs/xfs_message.h>
#include "xfs.h"
#include "ftp.h"
#include "dir.h"
#include "node.h"
#include "inum.h"
#include "ftpdir.h"
#include "ftproot.h"
#include "ftpfile.h"
#include "finderror.h"
#include "hash.h"
#include "ihash.h"
#include "parse_ls.h"

#undef USE_FTP_STAT

/*
 * ftpdir: a node representing a directory on a remote FTP server.
 */

/* Move these two lines into their own header file: */
extern int debugflag;

struct ftpdir_node {
	struct node		node;
	struct ftpdir_node	*parent;
	struct node		*root;
	char			*name;
	struct hash		*hash;
	struct ihash		*ihash;
};

static void forget(struct ftpdir_node *, const char *);
static ino_t get_inum(struct ftpdir_node *, const char *);

int
statname(root, name, st, uidp, gidp, linkp)
	struct node *root;
	const char *name;
	struct stat *st;
	char **uidp, **gidp, **linkp;
{
	struct ftp_ctx *ctx;
	struct ftp_response resp;
	char *line;
	int isowner;
	char *e;
#if !defined(USE_FTP_STAT)
	int fd;
	int len, left;
	char buf[4096];
#endif

	/* XXX Cannot yet use MDTM because of broken firewall proxy */

	ctx = ftproot_ctx(root);
#if defined(USE_FTP_STAT)
	ftp_sendcmd(ctx, &resp, "STAT -d %s", name);
	line = resp.msg;
	/* Skip the response line */
	while (*line && *line != '\n')
		line++;
	if (*line) line++;
	e = line;
	while (*e && *e != '\n')
		e++;
	*e = '\0';
	while (*line == ' ' || *line == '\t')
		line++;
#else
	if ((fd = ftp_ntransfercmd(ctx, &resp, "NLST -ld %s", name)) == -1)
		return -1;
	/* XXX only fetch one line */
	line = buf;
	left = sizeof buf - 1;
	while ((len = read(fd, line, left)) != -1) {
		if (len == 0)
			break;
		line += len;
		left -= len;
		if (!left) {
			/* XXX abort? */
			break;
		}
	}
	close(fd);
	if (len == -1)
		/* XXX reset? */
		return -1;
	len = line - buf;
	buf[len] = '\0';
	line = buf;
    next_line:
	e = line + strcspn(buf, "\r\n");
	*e = '\0';
	if (e == line && e != &buf[len]) {
		line = e + 1;
		goto next_line;
	}
	if (debugflag)
		fprintf(stderr, "=> %s%s\n", line, e == line ? " [empty]":"");
	if (ftp_getresp(ctx, &resp) == -1)
		return -1;
#endif
	if (parse_ls(line, e - line, st, NULL, uidp, gidp, linkp) == -1)
		return finderror(line, ENOENT);

	if (uidp) {
		/* If its not our name, we use root for the uid/gid. */
		isowner = ftproot_isusername(root, *uidp);
		st->st_uid = isowner ? root->stat.st_uid : 0;
		st->st_gid = isowner ? root->stat.st_gid : 0;
	}

	return 0;
}

static int
ftpdir_getattr(fdn)
	struct ftpdir_node *fdn;
{
	char *uid;
	struct node *cwd;
	const char *nam;

	if (fdn->parent) {
		cwd = &fdn->parent->node;
		nam = fdn->name;
	} else {
		cwd = &fdn->node;
		nam = ".";
	}

	ftproot_lock(fdn->root);
	if (ftpdir_cwd(cwd) == -1) {
		ftproot_reset(fdn->root);
		ftproot_unlock(fdn->root);
		return -1;
	}
	ftproot_alive(fdn->root);
	if (statname(fdn->root, nam, &fdn->node.stat, &uid, NULL, NULL) == -1) {
		/* ftproot_reset(fdn->root);*/
		ftproot_unlock(fdn->root);
		return -1;
	}

	ftproot_unlock(fdn->root);

	if (!S_ISDIR(fdn->node.stat.st_mode)) {
		warnx("%s: parse_ls did not detect a directory", fdn->name);
		errno = EIO;
		return -1;
	}
	return 0;
}

static int
ftpdir_putattr(fdn, st)
	struct ftpdir_node *fdn;
	struct stat *st;
{
	struct ftp_response resp;

	if ((st->st_mode & ALLPERMS) != (fdn->node.stat.st_mode & ALLPERMS)) {
		ftproot_lock(fdn->root);
		if (ftpdir_cwd(&fdn->node) == -1) {
			ftproot_reset(fdn->root);
			ftproot_unlock(fdn->root);
			return -1;
		}
		if (ftp_sendcmd(ftproot_ctx(fdn->root), &resp, 
		    "SITE CHMOD 0%o %s.", st->st_mode & ALLPERMS, fdn->name) 
		    == -1) {
			ftproot_reset(fdn->root);
			ftproot_unlock(fdn->root);
			return -1;
		}
		ftproot_alive(fdn->root);
		ftproot_unlock(fdn->root);
		if (ftpdir_getattr(fdn) == -1)
			return -1;
	}
	return 0;
}

static int
ftpdir_getdata(fdn)
	struct ftpdir_node *fdn;
{
	struct ftp_response resp;
	int fd, rfd;
	FILE *rfp;
	char *line;
	size_t len;
	struct dirent de;
	char *uid;
	char *name;
	struct stat st;

	fd = node_open_file(&fdn->node, O_TRUNC|O_WRONLY);

	ftproot_lock(fdn->root);
	if (ftpdir_cwd(&fdn->node) == -1) {
		ftproot_reset(fdn->root);
		ftproot_unlock(fdn->root);
		return -1;
	}
	ftproot_alive(fdn->root);
	rfd = ftp_ntransfercmd(ftproot_ctx(fdn->root), &resp, "NLST -la");
	if (rfd == -1) {
		close(fd);
		ftproot_reset(fdn->root);
		ftproot_unlock(fdn->root);
		return -1;
	}
	rfp = fdopen(rfd, "r");

	DIRENT_SET(&de, &fdn->node, ".");
	dir_write(fd, &de);

	if (fdn->parent)
		DIRENT_SET(&de, &fdn->parent->node, "..");
	else
		DIRENT_SET(&de, &fdn->node, "..");
	dir_write(fd, &de);

	while ((line = fgetln(rfp, &len)) != NULL) {
		while (len && (line[len - 1] == '\r' || line[len - 1] == '\n'))
			len --;
		if (debugflag)
			fprintf(stderr, "=> %.*s\n", (int)len, line);
		if (len >= 6 && memcmp(line, "total ", 6) == 0)
			continue;
		if (parse_ls(line, len, &st, &name, &uid, NULL, NULL) == 0) {
			if (debugflag)
				fprintf(stderr,
				    "name=%s uid=%s sz=%qd\n",
				    name, uid, st.st_size);
			if (strcmp(name, ".") == 0 ||
			    strcmp(name, "..") == 0)
				continue;
			if (!S_ISDIR(st.st_mode) && 
			    !S_ISREG(st.st_mode) &&
			    !S_ISLNK(st.st_mode))
				continue;
			de.d_fileno = get_inum(fdn, name);
			de.d_type = IFTODT(st.st_mode);
			de.d_namlen = strlen(name);
			bzero(de.d_name, sizeof de.d_name);
			strlcpy(de.d_name, name, sizeof de.d_name);
			de.d_reclen = sizeof de;
			if (debugflag)
				fprintf(stderr, 
				    "[reclen=%x fileno=%x type=%x "
				    "namlen=%x name=%.*s]\n",
				    de.d_reclen, de.d_fileno, de.d_type,
				    de.d_namlen, de.d_namlen, de.d_name);
			dir_write(fd, &de);
		}
	}
	dir_close(fd);
	fclose(rfp);
	if (ftp_getresp(ftproot_ctx(fdn->root), &resp) == -1) {
		ftproot_reset(fdn->root);
		ftproot_unlock(fdn->root);
		return -1;
	}
	ftproot_unlock(fdn->root);

	fstat(fd, &st);
	fdn->node.stat.st_size = st.st_size;
	fdn->node.tokens &= ~XFS_ATTR_VALID;

	return 0;
}

static int
ftpdir_putdata(fdn)
	struct ftpdir_node *fdn;
{
	return 0;
}

static int
ftpdir_create(fdn, name, st)
	struct ftpdir_node *fdn;
	const char *name;
	struct stat *st;
{
	struct ftp_response resp;
	int fd;

	if (S_ISDIR(st->st_mode)) {
		ftproot_lock(fdn->root);
		if (ftp_sendcmd(ftproot_ctx(fdn->root), &resp, "MKD %s%s", 
		    fdn->name, name) == -1) {
			ftproot_reset(fdn->root);
			ftproot_unlock(fdn->root);
			return -1;
		}
		ftproot_alive(fdn->root);
		ftproot_unlock(fdn->root);
		/* ftpdir_getdata(fdn); */
	} else if (S_ISREG(st->st_mode)) {
		ftproot_lock(fdn->root);
		fd = ftp_ntransfercmd(ftproot_ctx(fdn->root), &resp,
		    "STOR %s%s", fdn->name, name);
		if (fd < 0) {
			ftproot_reset(fdn->root);
			ftproot_unlock(fdn->root);
			return -1;
		}
		/* Send 0 bytes */
		close(fd);
		if (ftp_getresp(ftproot_ctx(fdn->root), &resp) == -1) {
			ftproot_reset(fdn->root);
			ftproot_unlock(fdn->root);
			return -1;
		}
		ftproot_alive(fdn->root);
		ftproot_unlock(fdn->root);
	} else {
		errno = ENODEV;
		return -1;
	}
	forget(fdn, name);
	return 0;
}

static int
ftpdir_destroy(fdn, name)
	struct ftpdir_node *fdn;
	const char *name;
{
	struct ftp_response resp;
	struct node *n;
	int ret;
	int isdir, isreg;

	if (hash_get(fdn->hash, name, &n))
		n = NULL;
	if (n) {
		isdir = S_ISDIR(n->stat.st_mode);
		isreg = !isdir;
		forget(fdn, name);
	} else
		isdir = isreg = 1;

	ret = 0;
	if (isdir) {
		ftproot_lock(fdn->root);
		ret = ftp_sendcmd(ftproot_ctx(fdn->root), &resp, "RMD %s%s", 
		    fdn->name, name);
		if (ret == 0)
			ftproot_alive(fdn->root);
		else
			ftproot_reset(fdn->root);
		ftproot_unlock(fdn->root);
	}
	if (ret && isreg) {
		ftproot_lock(fdn->root);
		ret = ftp_sendcmd(ftproot_ctx(fdn->root), &resp, "DELE %s%s", 
		    fdn->name, name);
		if (ret == 0)
			ftproot_alive(fdn->root);
		else
			ftproot_reset(fdn->root);
		ftproot_unlock(fdn->root);
	}
	return ret;
}

static struct node *
ftpdir_lookup(fdn, name, cred)
	struct ftpdir_node *fdn;
	const char *name;
	struct ncred *cred;
{
	struct stat st;
	struct node *n;
	char *uid;
	char *link;

	if (strcmp(name, ".") == 0)
		return &fdn->node;
	if (strcmp(name, "..") == 0) {
		if (fdn->parent)
			return &fdn->parent->node;
		else
			return &fdn->node;
	}

	if (hash_get(fdn->hash, name, &n) == 0)
		return n;

	ftproot_lock(fdn->root);
	if (ftpdir_cwd(&fdn->node) == -1) {
		ftproot_reset(fdn->root);
		ftproot_unlock(fdn->root);
		return NULL;
	}
	if (statname(fdn->root, name, &st, &uid, NULL, &link) == -1) {
		ftproot_reset(fdn->root);
		ftproot_unlock(fdn->root);
		return NULL;
	}
	ftproot_alive(fdn->root);
	ftproot_unlock(fdn->root);


	st.st_ino = get_inum(fdn, name);
	st.st_uid = 0;
	if (S_ISDIR(st.st_mode)) 
		n = ftpdir_new(&fdn->node, fdn->root, name, 
			st.st_ino);
	else 
		n = ftpfile_new(fdn->root, &fdn->node, name, &st, link);

    gotit:
	hash_add(fdn->hash, name, n);
	return n;
}

static int
ftpdir_rename(fdn_from, name_from, n_to, name_to)
	struct ftpdir_node *fdn_from;
	const char *name_from;
	struct node *n_to;
	const char *name_to;
{
	struct ftp_response resp;
	struct ftpdir_node *fdn_to = (struct ftpdir_node *)n_to;

	/* Use the real root directory, if the ftp root node was given */
	if (n_to == fdn_from->root)
		n_to = ftproot_root(n_to);

	/* Only support renaming to other FTP directories */
	if (n_to->class != &ftpdir_class) {
		errno = EXDEV;
		return -1;
	}

	/* Only allow renaming within this ftp context */
	if (fdn_to->root != fdn_from->root) {
		errno = EXDEV;
		return -1;
	}

	ftproot_lock(fdn_from->root);
	ftp_sendcmd(ftproot_ctx(fdn_from->root), &resp, "RNFR %s%s", 
		fdn_from->name, name_from);
	ftp_sendcmd(ftproot_ctx(fdn_to->root), &resp, "RNTO %s%s", 
		fdn_to->name, name_to);
	ftproot_alive(fdn_from->root);
	ftproot_unlock(fdn_from->root);

	forget(fdn_from, name_from);
	forget(fdn_to, name_to);

	return 0;
}

static void
ftpdir_dealloc(fdn)
	struct ftpdir_node *fdn;
{
	char name[MAXNAMLEN];
	struct node *n;

	if (fdn->hash) {
		if (debugflag) {
			hash_first(fdn->hash); 
			while (hash_next(fdn->hash, name, sizeof name, &n) == 0) {
				fprintf(stderr, "hash \"%s\" -> ", name);
				node_repr(n, stderr);
				fprintf(stderr, "\n");
			}
		}
				
		hash_first(fdn->hash); 
		while (hash_next(fdn->hash, name, sizeof name, &n) == 0) {
			if (strcmp(".", name) == 0 || strcmp("..", name) == 0)
				continue;
			node_invalidnode(n);
		}
		hash_free(fdn->hash);
	}
	if (fdn->ihash)
		/* XXX may need to reclaim all the inums?? */
		ihash_free(fdn->ihash);
	if (fdn->name)
		free(fdn->name);
	node_deinit(&fdn->node);
	free(fdn);
}

static void
ftpdir_repr(fdn, f)
	struct ftpdir_node *fdn;
	FILE *f;
{
	fprintf(f, "<ftpdir %s>", fdn->name ? fdn->name : "-");
}

struct node_class ftpdir_class = {
	ftpdir_getattr,	/* getattr */
	ftpdir_putattr,	/* putattr */
	ftpdir_getdata,	/* getdata */
	ftpdir_putdata,	/* putdata */
	ftpdir_create,	/* create  */
	ftpdir_destroy,	/* destroy */
	ftpdir_lookup,	/* lookup  */
	ftpdir_rename,	/* rename  */
	ftpdir_dealloc,	/* dealloc */
	ftpdir_repr	/* repr */
};

struct node *
ftpdir_new(parent, root, name, ino)
	struct node *parent;
	struct node *root;
	const char *name;
	ino_t ino;
{
	struct ftpdir_node *fdn;
    	struct ftpdir_node *fdparent = (struct ftpdir_node *)parent;
	int namelen, pnamelen;

	if (debugflag) {
		fprintf(stderr, "ftpdir_new(parent=");
		node_repr(parent, stderr);
		fprintf(stderr, ", root=");
		node_repr(root, stderr);
		fprintf(stderr, ", name=\"%s\"\n", name);
	}

	if (parent && parent->class != &ftpdir_class)
		err(1, "ftpdir_new: parent is not ftpdir");
	if (root->class != &ftproot_class)
		err(1, "ftpdir_new: root is not ftproot");

	fdn = (struct ftpdir_node *)malloc(sizeof (struct ftpdir_node));
	if (fdn == NULL)
		return NULL;
	node_init(&fdn->node, &ftpdir_class, root->ctx);

	fdn->node.stat.st_ino = ino;
	fdn->name = NULL;
	fdn->root = root;
	fdn->parent = fdparent;
	if ((fdn->hash = hash_new()) == NULL)
		goto err;
	if ((fdn->ihash = ihash_new()) == NULL)
		goto err;

	/* Compute the name of this directory on the server with trailing / */
	namelen = strlen(name);
	pnamelen = fdn->parent ? strlen(fdn->parent->name) : 0;
	if ((fdn->name = malloc(namelen + pnamelen + 2)) == NULL)
		goto err;
	if (pnamelen)
		memcpy(fdn->name, fdn->parent->name, pnamelen);
	memcpy(fdn->name + pnamelen, name, namelen);
	fdn->name[pnamelen + namelen] = '/';
	fdn->name[pnamelen + namelen + 1] = '\0';

	if (ftpdir_getattr(fdn) == -1) 
		goto err;

	return (struct node *)fdn;
    err:
	ftpdir_dealloc(fdn);
	return NULL;
}

int
ftpdir_cwd(n)
	struct node *n;
{
	struct ftpdir_node *fdn = (struct ftpdir_node *)n;
	struct ftp_response resp;

	ftp_sendcmd(ftproot_ctx(fdn->root), &resp, "CWD %s", fdn->name);
	return 0;
}

/* Forget a child, and update a changed directory listing. */
static void
forget(fdn, name)
	struct ftpdir_node *fdn;
	const char *name;
{
	struct node *n;

	if (strcmp(name, ".") == 0 || strcmp(name, "..") == 0)
		return;
	if (hash_get(fdn->hash, name, &n) == 0) {
		hash_del(fdn->hash, name);
		ihash_del(fdn->ihash, name);
		node_invalidnode(n);
	}
	fdn->node.tokens &= ~(XFS_DATA_R | XFS_DATA_W);
	node_installattr(&fdn->node);
}

static ino_t
get_inum(fdn, name)
	struct ftpdir_node *fdn;
	const char *name;
{
	ino_t inum;

	if (strcmp(name, ".") == 0)
		return fdn->node.stat.st_ino;
	if (strcmp(name, "..") == 0) {
		if (fdn->parent)
			return fdn->parent->node.stat.st_ino;
		return fdn->node.stat.st_ino;
	}

	if (ihash_get(fdn->ihash, name, &inum) == 0)
		return inum;

	inum = inum_new();
	ihash_add(fdn->ihash, name, inum);

	return inum;
}
