/* $Id: ftpfile.c,v 1.5 2000/02/29 05:25:22 leonard Exp $ */
/* Copyright David Leonard, 2000. All rights reserved. */
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <stdarg.h>
#include <err.h>
#include <fcntl.h>
#include <errno.h>
#include <string.h>
#include <sys/socket.h>
#include <sys/stat.h>
#include <sys/dirent.h>
#include <xfs/xfs_message.h>
#include "xfs.h"
#include "node.h"
#include "ftp.h"
#include "ftpdir.h"
#include "ftproot.h"
#include "parse_ls.h"

extern int debugflag;

/*
 * ftpfile file: s/ftpfile/something/g
 *		  s/ffn/sn/g;
 * extern struct node_class ftpfile_class;
 */

struct ftpfile_node {
	struct node		node;
	struct node		*parent;
	struct node		*root;
	char			name[MAXNAMLEN];
};

static int
ftpfile_getattr(ffn)
	struct ftpfile_node *ffn;
{
	char *uid;

	ftproot_lock(ffn->root);
	if (ftpdir_cwd(ffn->parent) == -1) {
		ftproot_reset(ffn->root);
		ftproot_unlock(ffn->root);
		return -1;
	}
	ftproot_alive(ffn->root);

	if (statname(ffn->root, ffn->name, 
	    &ffn->node.stat, &uid, NULL, NULL) == -1) {
		ftproot_reset(ffn->root);
		ftproot_unlock(ffn->root);
		return -1;
	}
	ftproot_unlock(ffn->root);

	if (S_ISDIR(ffn->node.stat.st_mode)) {
		warnx("%s: parse_ls detected a directory", ffn->name);
		errno = EIO;
		return -1;
	}

	return 0;
}

static int
ftpfile_putattr(ffn, st)
	struct ftpfile_node *ffn;
	struct stat *st;
{
	struct ftp_response resp;

	if ((st->st_mode & ALLPERMS) != (ffn->node.stat.st_mode & ALLPERMS)) {
		ftproot_lock(ffn->root);
		if (ftpdir_cwd(ffn->parent) == -1) {
			ftproot_reset(ffn->root);
			ftproot_unlock(ffn->root);
			return -1;
		}
		ftproot_alive(ffn->root);
		if (ftp_sendcmd(ftproot_ctx(ffn->root), &resp,
		    "SITE CHMOD 0%o %s.", st->st_mode & ALLPERMS, ffn->name)
		    == -1) {
			ftproot_reset(ffn->root);
			ftproot_unlock(ffn->root);
			return -1;
		}
		ftproot_unlock(ffn->root);
		if (ftpfile_getattr(ffn) == -1)
			return -1;
	}
	return 0;
}

static int
ftpfile_getdata(ffn)
	struct ftpfile_node *ffn;
{
	struct ftp_response resp;
	int fd, rfd;
	char buf[8192];
	ssize_t len;

	if (S_ISLNK(ffn->node.stat.st_mode)) 
		return 0;

	if (!S_ISREG(ffn->node.stat.st_mode)) {
		errno = ENODEV;
		return -1;
	}

	if ((fd = node_open_file(&ffn->node, O_TRUNC|O_WRONLY)) == -1)
		return -1;

	ftproot_lock(ffn->root);
	if (ftpdir_cwd(ffn->parent) == -1) {
		ftproot_reset(ffn->root);
		ftproot_unlock(ffn->root);
		close(fd);
		return -1;
	}
	ftproot_alive(ffn->root);
	if ((rfd = ftp_ntransfercmd(ftproot_ctx(ffn->root), &resp, 
	    "RETR %s", ffn->name)) == -1) {
		ftproot_reset(ffn->root);
		ftproot_unlock(ffn->root);
		close(fd);
		return -1;
	}

	while ((len = read(rfd, buf, sizeof buf)) > 0)
		if (write(fd, buf, len) != len)
			break;
	close(rfd);
	ftp_getresp(ftproot_ctx(ffn->root), &resp);
	ftproot_unlock(ffn->root);
	close(fd);
	return 0;
}

static int
ftpfile_putdata(ffn)
	struct ftpfile_node *ffn;
{
	struct ftp_response resp;
	int fd, wfd;
	char buf[8192];
	ssize_t len;

	if (!S_ISREG(ffn->node.stat.st_mode)) {
		errno = ENODEV;
		return -1;
	}

	if ((fd = node_open_file(&ffn->node, O_RDONLY)) == -1)
		return -1;
	ftproot_lock(ffn->root);
	if (ftpdir_cwd(ffn->parent) == -1) {
		close(fd);
		ftproot_reset(ffn->root);
		ftproot_unlock(ffn->root);
		return -1;
	}
	ftproot_alive(ffn->root);
	if ((wfd = ftp_ntransfercmd(ftproot_ctx(ffn->root), &resp, 
	    "STOR %s", ffn->name)) == -1) {
		close(fd);
		ftproot_reset(ffn->root);
		ftproot_unlock(ffn->root);
		return -1;
	}

	while ((len = read(fd, buf, sizeof buf)) > 0)
		if (write(wfd, buf, len) != len)
			break;
	close(wfd);
	ftp_getresp(ftproot_ctx(ffn->root), &resp);
	ftproot_unlock(ffn->root);
	close(fd);
	return 0;
}

static void
ftpfile_dealloc(ffn)
	struct ftpfile_node *ffn;
{
	node_deinit(&ffn->node);
	free(ffn);
}

static void
ftpfile_repr(ffn, f)
	struct ftpfile_node *ffn;
	FILE *f;
{

	fprintf(stderr, "<ftpfile %s>", ffn->name);
}

struct node_class ftpfile_class = {
	ftpfile_getattr,	/* getattr */
	ftpfile_putattr,	/* putattr */
	ftpfile_getdata,	/* getdata */
	ftpfile_putdata,	/* putdata */
	(create_t)enotdir,	/* create  */
	(destroy_t)enotdir,	/* destroy */
	(lookup_t)enotdir,	/* lookup  */
	(rename_t)enotdir,	/* rename  */
	ftpfile_dealloc,	/* dealloc */
	ftpfile_repr		/* repr    */
};

struct node *
ftpfile_new(root, parent, name, st, link)
	struct node *root;
	struct node *parent;
	const char *name;
	struct stat *st;
	const char *link;
{
	struct ftpfile_node *ffn;
	int fd;

	ffn = (struct ftpfile_node *)malloc(sizeof (struct ftpfile_node));
	if (ffn == NULL)
		return NULL;
	node_init(&ffn->node, &ftpfile_class, parent->ctx);
	ffn->node.stat = *st;
	ffn->root = root;
	ffn->parent = parent;
	if (S_ISLNK(st->st_mode)) {
		if (debugflag)
			fprintf(stderr, "@@ Symlink %s to %s\n", name, link);
		if (node_alloc_file(&ffn->node) == -1) {
			ftpfile_dealloc(ffn);
			return NULL;
		}
		fd = node_open_file(&ffn->node, O_TRUNC | O_WRONLY);
		if (fd < 0) {
			ftpfile_dealloc(ffn);
			return NULL;
		}
		if (write(fd, link, strlen(link)) == -1) {
			close(fd);
			ftpfile_dealloc(ffn);
			return NULL;
		}
		close(fd);
	}
	strncpy(ffn->name, name, sizeof ffn->name);
	return (struct node *)ffn;
}
