/* $Id: ftproot.c,v 1.5 2000/02/29 10:22:57 leonard Exp $ */
/* Copyright David Leonard, 2000. All rights reserved. */
#include <stdio.h>
#include <fcntl.h>
#include <errno.h>
#include <unistd.h>
#include <string.h>
#include <stdlib.h>
#include <stdarg.h>
#include <pthread.h>
#include <dirent.h>
#include <err.h>
#include <sys/stat.h>
#include <sys/socket.h>
#include <xfs/xfs_message.h>
#include "inum.h"
#include "ftp.h"
#include "dir.h"
#include "xfs.h"
#include "debug.h"
#include "node.h"
#include "ftproot.h"
#include "ftppasswd.h"
#include "ftpdir.h"

/*
 * ftproot: an FTP session.
 * This is a directory node containing the file ".passwd" (ftppasswd)
 * and is aggregated with the top level directory when the FTP session
 * is connected.
 *
 * write .passwd -- attempt to connect the FTP session
 * rm .passwd -- disconnect the FTP session
 * All other operations are either forwarded to the FTP root 
 * directory (ftpfile), or the 'Socket not connected' error is returned.
 */

struct ftproot_node {
	struct node		node;
	struct node 		*parent;
	struct node 		*passwd;
	struct node 		*root;
	struct ftp_ctx		*ftp;
	pthread_t		keepalive;	/* periodic NOOPs */
	pthread_mutex_t		lock;		/* Lock when using ftp */
	pthread_cond_t		keptalive;	/* signal when conn used */
};

static void ftproot_disconnect(struct ftproot_node *);
static void *keepalive_thread(void *);

static int
ftproot_getattr(frn)
	struct ftproot_node *frn;
{
	if (frn->root)
		frn->node.tokens = frn->root->tokens;
	return 0;
}

static int
ftproot_getdata(frn)
	struct ftproot_node *frn;
{
	int fd;
	int rfd;
	struct dirent de;
	char buf[1024];		/* XXX blocksize */
	char *bp;

	if (frn->root) {
		if (!frn->root->file_valid)
			if (node_alloc_file(frn->root) == -1)
				return -1;
		if (node_getdata(frn->root) == -1)
			return -1;
	}

	if ((fd = node_open_file(&frn->node, O_WRONLY)) == -1)
		return -1;

	DIRENT_SET(&de, &frn->node, ".");
	dir_write(fd, &de);

	DIRENT_SET(&de, frn->parent, "..");
	dir_write(fd, &de);

	DIRENT_SET(&de, frn->passwd, ".passwd");
	dir_write(fd, &de);

	/* If open, copy the underlying directory's listing to us */
	if (frn->root) {
		if ((rfd = node_open_file(frn->root, O_RDONLY)) == -1) {
			close(fd);
			return -1;
		}

		while (read(rfd, buf, sizeof buf) == sizeof buf) {
		    bp = buf;
		    while (bp < buf + sizeof buf) {
			memcpy(&de, bp, sizeof de);
			if (debugflag)
				fprintf(stderr, "@@ %u %.*s\n",
					de.d_fileno, de.d_namlen, de.d_name);
			if (de.d_reclen == 0)
				break;
			bp = bp + de.d_reclen;
			if (DIRENT_NAMEEQ(&de, "."))
				continue;
			if (DIRENT_NAMEEQ(&de, ".."))
				continue;
			if (DIRENT_NAMEEQ(&de, ".passwd"))
				continue;
			dir_write(fd, &de);
		    }
		}
		close(rfd);
	}

	dir_close(fd);

	return 0;
}

static int
ftproot_create(frn, name, st)
	struct ftproot_node *frn;
	const char *name;
	struct stat *st;
{
	if (frn->root) 
		return node_create(frn->root, name, st);
	errno = ENOTCONN;
	return -1;
}

static int
ftproot_destroy(frn, name)
	struct ftproot_node *frn;
	const char *name;
{

	/*
	 * When .passwd is removed, the connection is closed, but
	 * the .passwd file remains. All other files then disappear,
	 * so this is really the converse of 'rm .passwd'!
	 */
	if (strcmp(name, ".passwd") == 0) {
		if (frn->root == NULL) {
			errno = ENOTCONN;
			return -1;
		}
		ftproot_lock(&frn->node);
		if (frn->passwd)
			ftppasswd_invalidate(frn->passwd);
		ftproot_disconnect(frn);
		ftproot_unlock(&frn->node);
		return 0;
	}

	if (frn->root)
		return node_destroy(frn->root, name);

	errno = ENOENT;
	return -1;
}

static struct node *
ftproot_lookup(frn, name, cred)
	struct ftproot_node *frn;
	const char *name;
	struct ncred *cred;
{
	if (strcmp(name, ".passwd") == 0)
		return frn->passwd;
	if (strcmp(name, ".") == 0)
		return &frn->node;
	if (strcmp(name, "..") == 0)
		return frn->parent;
	if (frn->root)
		return node_lookup(frn->root, name, cred);
	errno = ENOENT;
	return NULL;
}

static int
ftproot_rename(frn, name_from, n_to, name_to)
	struct ftproot_node *frn;
	const char *name_from;
	struct node *n_to;
	const char *name_to;
{
	if (frn->root)
		return node_rename(frn->root, name_from,
			n_to, name_to);
	errno = ENODEV;
	return -1;
}

static void
ftproot_dealloc(frn)
	struct ftproot_node *frn;
{
	void *res;

	ftproot_lock(&frn->node);
	if(frn->root)
		ftproot_disconnect(frn);
	node_invalidnode(frn->passwd);
	if (frn->ftp)
		ftp_free_ctx(frn->ftp);
	pthread_cond_signal(&frn->keptalive);
	ftproot_unlock(&frn->node);
	pthread_join(frn->keepalive, &res);
	frn->ftp = NULL;
	inum_reclaim(frn->node.stat.st_ino);
	node_deinit(&frn->node);
	free(frn);
}

static void
ftproot_repr(frn, f)
	struct ftproot_node *frn;
	FILE *f;
{
	fprintf(f, "<ftproot user=%s host=%s%s>",
		frn->ftp ? frn->ftp->username : "-",
		frn->ftp ? frn->ftp->hostname : "-",
		frn->root ? " hasroot" : "");
}

struct node_class ftproot_class = {
	ftproot_getattr,
	(putattr_t)eperm,
	ftproot_getdata,
	(putdata_t)eperm,
	ftproot_create,
	ftproot_destroy,
	ftproot_lookup,
	ftproot_rename,
	ftproot_dealloc,
	ftproot_repr
};

/* Create a new FTP root */
struct node *
ftproot_new(parent, hostname, username, st, passive)
	struct node *parent;
	const char *hostname;
	const char *username;
	struct stat *st;
	int passive;
{
	struct ftproot_node *frn;

	frn = (struct ftproot_node *)malloc(sizeof (struct ftproot_node));
	if (frn == NULL)
		return NULL;
	node_init(&frn->node, &ftproot_class, parent->ctx);
	frn->parent = parent->class ? parent : &frn->node;
	frn->node.stat.st_nlink = 3;
	frn->node.stat.st_ino = inum_new();

	/* Copy the owner and mode from the credentials */
	frn->node.stat.st_uid = st->st_uid;
	frn->node.stat.st_gid = st->st_gid;
	frn->node.stat.st_mode = (st->st_mode & ALLPERMS) | S_IFDIR;

	/* The .passwd file is write-only by the owner */
	frn->passwd = ftppasswd_new(&frn->node);
	frn->passwd->stat.st_uid = st->st_uid;

	/* Create an FTP context, no password, not connected. */
	if (debugflag)
		fprintf(stderr, "Creating %s ftp context\n",
		    passive ? "PASSIVE" : "ACTIVE");
	frn->ftp = ftp_new_ctx(hostname, username, "", passive);
	frn->root = NULL;

	pthread_mutex_init(&frn->lock, NULL);
	pthread_cond_init(&frn->keptalive, NULL);
	pthread_create(&frn->keepalive, NULL, keepalive_thread, frn);

	return &frn->node;
}

/* Accept a password and try to make connection */
int
ftproot_connect(n)
	struct node *n;
{
	struct ftproot_node *frn = (struct ftproot_node *)n;
	int len;

	if (frn->root) {
		errno = EISCONN;
		return -1;
	}

	len = ftppasswd_getpasswd(frn->passwd, frn->ftp->password, 
	    sizeof frn->ftp->password);
	if (len == -1)
		return -1;

	ftproot_alive(&frn->node);
	if (ftp_connect(frn->ftp) == -1)
		return -1;

	frn->root = ftpdir_new(NULL, &frn->node, "", 0);
	if (frn->root == NULL) {
		int errsave = errno;
		(void)ftp_quit(frn->ftp);
		errno = errsave;
		return -1;
	}
	frn->node.tokens &= ~(XFS_DATA_R | XFS_DATA_W);
	node_installattr(&frn->node);
	return 0;
}

void
ftproot_lock(n)
	struct node *n;
{
	struct ftproot_node *frn = (struct ftproot_node *)n;

	pthread_mutex_lock(&frn->lock);
}

void
ftproot_unlock(n)
	struct node *n;
{
	struct ftproot_node *frn = (struct ftproot_node *)n;
	int errsave = errno;

	pthread_mutex_unlock(&frn->lock);
	errno = errsave;
}

/* Return true if this node is logged in */
int
ftproot_isconnected(n)
	struct node *n;
{
	struct ftproot_node *frn = (struct ftproot_node *)n;

	return (frn->root != NULL);
}

/* Return 0 if the connection could be ensured valid */
int
ftproot_ensure_connected(n)
	struct node *n;
{
	struct ftp_response resp;
	struct ftproot_node *frn = (struct ftproot_node *)n;
	int fd;
	int len;
	char buf[1024];

	if (frn->root == NULL) {
		errno = ENOTCONN;
		return -1;
	}

	/* Test the connection with a NOOP */
	if (ftp_sendcmd(frn->ftp, &resp, "NOOP") == -1)
		if (ftproot_reset(n) == -1)
			return -1;

	/* XXX some FTP servers want a short transfer. */
	if ((fd = ftp_ntransfercmd(frn->ftp, &resp, "NLST -ld /")) == -1)
		return -1;
	while ((len = read(fd, buf, sizeof buf)) > 0)
		;
	close(fd);
	if (len == -1)
		return ftproot_reset(n);
	ftp_getresp(frn->ftp, &resp);

	return 0;
}

/* XXX needs locking */
int
ftproot_reset(n)
	struct node *n;
{
	struct ftproot_node *frn = (struct ftproot_node *)n;
	struct ftp_ctx *octx;
	int len;

	octx = frn->ftp;
	frn->ftp = ftp_new_ctx(octx->hostname, octx->username,
	    octx->password, octx->passive);
	ftp_free_ctx(octx);
	if (debugflag) {
		fprintf(stderr, "* Reconnecting");
		node_repr(n, stderr);
		fprintf(stderr, "\n");
	}

	len = ftppasswd_getpasswd(frn->passwd, frn->ftp->password, 
	    sizeof frn->ftp->password);
	if (len == -1)
		return -1;

	ftproot_alive(&frn->node);
	if (ftp_connect(frn->ftp) == -1)
		return -1;

	return 0;
}

static void
ftproot_disconnect(frn)
	struct ftproot_node *frn;
{

	(void)ftp_quit(frn->ftp);
	if (frn->root) {
		node_invalidnode(frn->root);
		frn->node.tokens &= ~(XFS_DATA_R | XFS_DATA_W);
		node_installattr(&frn->node);
		frn->root = NULL;
	}
}

/* Return true if the string given is the username logged in as */
int
ftproot_isusername(n, s)
	struct node *n;
	const char *s;
{
	struct ftproot_node *frn = (struct ftproot_node *)n;
	int len = strlen(s);
	const char *username = frn->ftp->username;

	return memcmp(s, username, len) == 0 &&
	    (username[len] == '\0' || username[len] == '@');
}

struct ftp_ctx *
ftproot_ctx(n)
	struct node *n;
{
	struct ftproot_node *frn = (struct ftproot_node *)n;

	return frn->ftp;
}

struct node *
ftproot_root(n)
	struct node *n;
{
	struct ftproot_node *frn = (struct ftproot_node *)n;

	return frn->root;
}

void
ftproot_alive(n)
	struct node *n;
{
	struct ftproot_node *frn = (struct ftproot_node *)n;

	pthread_cond_signal(&frn->keptalive);
}

/* This thread periodically calls ftproot_ensure_connected */
static void *
keepalive_thread(arg)
	void *arg;
{
	struct ftproot_node *frn = (struct ftproot_node *)arg;
	struct timespec wakeup;
	int err;

	ftproot_lock(&frn->node);
	while (1) {
		wakeup.tv_sec = time(0) + 120;	/* Two minutes */
		wakeup.tv_nsec = 0;
		err = pthread_cond_timedwait(&frn->keptalive, &frn->lock,
			&wakeup);
		if (debugflag) {
			fprintf(stderr, "* ");
			node_repr(&frn->node, stderr);
			fprintf(stderr, " keepalive woken up\n");
		}

		if (err == 0)		/* someone signalled us */
			continue;
		if (err == ETIMEDOUT) {
			if (ftproot_ensure_connected(&frn->node) == -1) {
				ftproot_disconnect(frn);
				ftproot_reset(&frn->node);
			}
			continue;
		}
		errx(1, "keepalive_thread: %s", strerror(err));
	}
	ftproot_unlock(&frn->node);
	return NULL;
}
