/* $Id: node.c,v 1.3 2000/02/29 04:15:49 leonard Exp $ */
/* Copyright David Leonard, 2000. All rights reserved. */
#include <unistd.h>
#include <fcntl.h>
#include <stdio.h>
#include <err.h>
#include <errno.h>
#include <string.h>
#include <sys/dirent.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <xfs/xfs_message.h>

#include "xfs.h"
#include "node.h"

/*
 * General node caching operations, control and methods
 */

extern int debugflag;

const char *cachebase = "/tmp/.cache.";

/* Generate the filename for the data file that backs a node */
static void
node_filename(n, path)
	struct node *n;
	char *path;
{

	snprintf(path, MAXPATHLEN, "%s%p", cachebase, n);
}


/* Allocate storage for the backing data file */
int
node_alloc_file(n)
	struct node *n;
{
	char name[MAXPATHLEN];
	int fd;

	if (n->file_valid) {
		warnx("node_alloc_file: already there");
		return 0;
	}

	node_filename(n, name);
	if ((fd = open(name, O_CREAT|O_TRUNC|O_WRONLY, 0600)) == -1)
		return -1;
	close(fd);
	if (xfs_fhget(name, 1, &n->file) == -1)
		return -1;
	n->file_valid = 1;
	n->file_external = 0;
	return 0;
}

/* Allocate storage for the backing data file */
int
node_external_file(n, name)
	struct node *n;
	char *name;
{
	int fd;

	if (n->file_valid)
		errx(1, "node_external_file: already valid");

	if ((fd = open(name, O_CREAT|O_RDONLY, 0600)) == -1)
		return -1;
	close(fd);
	if (xfs_fhget(name, 1, &n->file) == -1)
		return -1;
	n->file_valid = 1;
	n->file_external = 1;
	return 0;
}


/* Open the backing data file onto a file descriptor */
int
node_open_file(n, flags)
	struct node *n;
	int flags;
{

	if (!n->file_valid) {
		errno = EIO;
		return -1;
	}

	return xfs_fhopen(&n->file, flags);
}

/* Get the size of the backing data file */
off_t
node_getsize_file(n)
	struct node *n;
{
	int fd;
	struct stat st;

	if ((fd = node_open_file(n, O_RDONLY)) == -1)
		return -1;
	if (fstat(fd, &st) == -1) {
		close(fd);
		return -1;
	}
	close(fd);
	return st.st_size;
}

int
node_setsize_file(n, sz)
	struct node *n;
	off_t sz;
{
	int fd;
	if ((fd = node_open_file(n, O_RDONLY)) == -1)
		return -1;
	if (ftruncate(fd, sz) == -1) {
		close(fd);
		return -1;
	}
	close(fd);
	return 0;
}

/* Deallocate storage for a backing data file */
int
node_dealloc_file(n)
	struct node *n;
{
	char name[MAXPATHLEN];

	if (!n->file_valid) {
		warnx("node_dealloc_file: already gone");
		return 0;
	}
	n->file_valid = 0;
	if (!n->file_external) {
		node_filename(n, name);
		if (unlink(name) == -1)
			warn("unlink: %s", name);
	}
	return 0;
}

/* Convert an opaque XFS node handle into a node pointer */
struct node *
handle_to_node(h)
	const struct xfs_handle *h;
{
	struct node *n;

	memcpy(&n, h, sizeof n);
	return n;
}

/* Convert a node pointer into an opaque XFS node handle */
void
node_to_handle(n, h)
	const struct node *n;
	struct xfs_handle *h;
{

	bzero(h, sizeof *h);
	memcpy(h, &n, sizeof n);
}

/*
 * Basic initialisation of a node structure.
 * Usually called by the class's alloc() method
 */
void
node_init(n, c, ctx)
	struct node *n;
	struct node_class *c;
	struct xfs_ctx *ctx;
{

	n->class = c;
	n->ctx = ctx;
	/* n->parent = NULL;
	n->namelen = 0; */
	n->file_valid = 0;
	n->tokens = 0;
	bzero(&n->stat, sizeof n->stat);
	n->stat.st_mtime = 
	    n->stat.st_ctime = 
	    n->stat.st_atime = time(0);
}

void
node_deinit(n)
	struct node *n;
{
	if (n->file_valid)
		warnx("node_deinit: data still allocated");
}

/* Convenient functions that just return errors */

int
eperm(n)
	struct node *n;
{

	errno = EPERM;
	return -1;
}

int
eacces(n)
	struct node *n;
{

	errno = EACCES;
	return -1;
}

int
enotdir(n)
	struct node *n;
{

	errno = ENOTDIR;
	return -1;
}

int
enodev(n)
	struct node *n;
{

	errno = ENODEV;
	return -1;
}

int
enoop(n)
	struct node *n;
{

	return 0;
}

#if 1 /* !defined(__OPTIMIZE__) */

int
node_getattr(n)
	struct node *n;
{
	int ret;
	int errsave;

	if (debugflag) {
		fprintf(stderr, "getattr(");
		node_repr(n, stderr);
		fprintf(stderr, ")\n");
	}
	ret = (*n->class->getattr)(n);
	if (debugflag) {
		errsave = errno;
		fprintf(stderr, "getattr %d %s\n", ret,
			ret == -1 ? strerror(errsave) : "");
		errno = errsave;
	}
	return ret;
}

int
node_putattr(n, st)
	struct node *n;
	struct stat *st;
{
	int ret;
	int errsave;

	if (debugflag) {
		fprintf(stderr, "putattr(");
		node_repr(n, stderr);
		fprintf(stderr, ", <ino %u mode %03o uid %d gid %d sz %qu>)\n",
			st->st_ino, st->st_mode, st->st_uid,
			st->st_gid, st->st_size);
	}
	ret = (*n->class->putattr)(n, st);
	if (debugflag) {
		errsave = errno;
		fprintf(stderr, "putattr %d %s\n", ret,
			ret == -1 ? strerror(errsave) : "");
		errno = errsave;
	}
	return ret;
}

int
node_getdata(n)
	struct node *n;
{
	int ret;
	int errsave;

	if (debugflag) {
		fprintf(stderr, "getdata(");
		node_repr(n, stderr);
		fprintf(stderr, ")\n");
	}
	ret = (*n->class->getdata)(n);
	if (debugflag) {
		errsave = errno;
		fprintf(stderr, "getdata %d %s\n", ret,
			ret == -1 ? strerror(errsave) : "");
		errno = errsave;
	}
	return ret;
}

int
node_putdata(n)
	struct node *n;
{
	int ret;
	int errsave;

	if (debugflag) {
		fprintf(stderr, "putdata(");
		node_repr(n, stderr);
		fprintf(stderr, ")\n");
	}
	ret = (*n->class->putdata)(n);
	if (debugflag) {
		errsave = errno;
		fprintf(stderr, "putdata %d %s\n", ret,
			ret == -1 ? strerror(errsave) : "");
		errno = errsave;
	}
	return ret;
}

int
node_create(n, nm, st)
	struct node *n;
	const char *nm;
	struct stat *st;
{
	int ret;
	int errsave;

	if (debugflag) {
		fprintf(stderr, "create(");
		node_repr(n, stderr);
		fprintf(stderr, ", <ino %u mode %03o uid %d gid %d sz %qu>)",
			st->st_ino, st->st_mode, st->st_uid,
			st->st_gid, st->st_size);
		fprintf(stderr, ", \"%s\")\n", nm);
	}
	ret = (*n->class->create)(n, nm, st);
	if (debugflag) {
		errsave = errno;
		fprintf(stderr, "create %d %s\n", ret,
			ret == -1 ? strerror(errsave) : "");
		errno = errsave;
	}
	return ret;
}

int
node_destroy(n, nm)
	struct node *n;
	const char *nm;
{
	int ret;
	int errsave;

	if (debugflag) {
		fprintf(stderr, "destroy(");
		node_repr(n, stderr);
		fprintf(stderr, ", \"%s\")\n", nm);
	}
	ret = (*n->class->destroy)(n, nm);
	if (debugflag) {
		errsave = errno;
		fprintf(stderr, "destroy %d %s\n", ret,
			ret == -1 ? strerror(errsave) : "");
		errno = errsave;
	}
	return ret;
}

struct node *
node_lookup(n, nm, cred)
	struct node *n;
	const char *nm;
	struct ncred *cred;
{
	struct node *ret;
	int errsave;

	if (debugflag) {
		fprintf(stderr, "lookup(");
		node_repr(n, stderr);
		fprintf(stderr, ", \"%s\", c=%d)\n", nm, cred->uid);
	}
	ret = (*n->class->lookup)(n, nm, cred);
	if (debugflag) {
		errsave = errno;
		fprintf(stderr, "lookup ");
		if (ret == NULL)
			fprintf(stderr, "NULL %s", strerror(errsave));
		else
			node_repr(ret, stderr);
		fprintf(stderr, "\n");
		errno = errsave;
	}
	return ret;
}

int
node_rename(f, nf, t, nt)
	struct node *f, *t;
	const char *nf, *nt;
{
	int ret;
	int errsave;

	if (debugflag) {
		fprintf(stderr, "rename(");
		node_repr(f, stderr);
		fprintf(stderr, ", \"%s\", ", nf);
		node_repr(t, stderr);
		fprintf(stderr, ", \"%s\")\n", nt);
	}
	ret = (*f->class->rename)(f, nf, t, nt);
	if (debugflag) {
		errsave = errno;
		fprintf(stderr, "rename %d %s\n", ret,
			ret == -1 ? strerror(errsave) : "");
		errno = errsave;
	}
	return ret;
}

void
node_dealloc(n)
	struct node *n;
{
	if (debugflag) {
		fprintf(stderr, "dealloc(");
		node_repr(n, stderr);
		fprintf(stderr, ")\n");
	}
	(*n->class->dealloc)(n);
	if (debugflag)
		fprintf(stderr, "dealloc\n");
}

#endif
