/* $Id: node_install.c,v 1.2 2000/02/29 01:17:25 leonard Exp $ */
/* Copyright David Leonard, 2000. All rights reserved. */
#include <unistd.h>
#include <fcntl.h>
#include <stdio.h>
#include <err.h>
#include <errno.h>
#include <string.h>
#include <sys/dirent.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <xfs/xfs_message.h>

#include "xfs.h"
#include "node.h"
#include "debug.h"

/*
 * Provide calls back to the XFS layer
 */

static void
stat_to_xa(attr, st)
	struct xfs_attr *attr;
	struct stat *st;
{
	XA_SET_MODE(attr, st->st_mode);
	XA_SET_NLINK(attr, st->st_nlink);
	XA_SET_SIZE(attr, st->st_size);
	XA_SET_UID(attr, st->st_uid);
	XA_SET_GID(attr, st->st_gid);
	XA_SET_ATIME(attr, st->st_atime);
	XA_SET_MTIME(attr, st->st_mtime);
	XA_SET_CTIME(attr, st->st_ctime);
	XA_SET_FILEID(attr, st->st_ino);
	switch(st->st_mode & S_IFMT) {
	case S_IFREG:	XA_SET_TYPE(attr, XFS_FILE_REG); break;
	case S_IFDIR:	XA_SET_TYPE(attr, XFS_FILE_DIR); break;
	case S_IFBLK:	XA_SET_TYPE(attr, XFS_FILE_BLK); break;
	case S_IFCHR:	XA_SET_TYPE(attr, XFS_FILE_CHR); break;
	case S_IFLNK:	XA_SET_TYPE(attr, XFS_FILE_LNK); break;
	case S_IFSOCK:	XA_SET_TYPE(attr, XFS_FILE_SOCK); break;
	case S_IFIFO:	XA_SET_TYPE(attr, XFS_FILE_FIFO); break;
	default:	XA_SET_TYPE(attr, XFS_FILE_BAD);
			errx(1, "BAD filetype %03o", st->st_mode);
	}
}


static void
node_to_xfs_msg_node(xmn, n)
	struct xfs_msg_node *xmn;
	struct node *n;
{
	node_to_handle(n, &xmn->handle);
	xmn->tokens = n->tokens;
	if (n->file_valid)
		xmn->tokens |= XFS_DATA_VALID;
	stat_to_xa(&xmn->attr, &n->stat);
	bzero(&xmn->id, sizeof xmn->id);
	bzero(&xmn->rights, sizeof xmn->rights);
	xmn->anonrights = XFS_RIGHT_R | XFS_RIGHT_W | XFS_RIGHT_X;
}

/*
 * Must be called exactly once from getroot() handler.
 * node becomes the root node.
 */                     
int
node_installroot(root)
	struct node *root;
{
	struct xfs_msg_node xmn;

	if (debugflag) {
		fprintf(stderr, "installroot(");
		node_repr(root, stderr);
		fprintf(stderr, ")\n");
	}

	node_to_xfs_msg_node(&xmn, root);
	return (*root->ctx->x_kops->installroot)(root->ctx, &xmn);
}

/*                      
 * Inserts the node named 'name' into the directory 'parent_handle'
 * in the cache. Must only be done exactly once.
 */
int
node_installnode(parent, child, name)
	struct node *parent, *child;
	const char *name;
{
	struct xfs_msg_node xmn;
	xfs_handle ph;

	if (debugflag) {
		fprintf(stderr, "installnode(parent=");
		node_repr(parent, stderr);
		fprintf(stderr, ", child=");
		node_repr(child, stderr);
		fprintf(stderr, ", name=\"%s\")\n", name);
	}

        node_to_xfs_msg_node(&xmn, child);
	node_to_handle(parent, &ph);
	return (*parent->ctx->x_kops->installnode)(parent->ctx, &ph, 
	    name, &xmn);
}

/*                      
 * Updates the pre-installed cache data with 'node'
 * May not always work since the kernel cache may have dropped the node.
 */
int
node_installattr(node)
	struct node *node;
{
	struct xfs_msg_node xmn;

	if (debugflag) {
		fprintf(stderr, "installattr(node=");
		node_repr(node, stderr);
		fprintf(stderr, ")\n");
	}
        node_to_xfs_msg_node(&xmn, node);
	return (*node->ctx->x_kops->installattr)(node->ctx, &xmn);
}

/*
 * Install data for node.handle.
 * cache_handle contains vnode information for openfh()/getfh()
 * cache_name is relative to the wd of the process.
 * This also updates attributes.
 * if (flag & XFS_INVALID_DNLC) then all cache info for the (directory) node
 * is flushed.                  
 */
int
node_installdata(node, inval)
	struct node *node;
	int inval;
{
	struct xfs_msg_node xmn;
	
	if (debugflag) {
		fprintf(stderr, "installdata(node=");
		node_repr(node, stderr);
		fprintf(stderr, ", inval=%d)\n", inval);
	}
	node_to_xfs_msg_node(&xmn, node);
	return (*node->ctx->x_kops->installdata)(node->ctx, &xmn, "",
		&node->file, inval ? XFS_INVALID_DNLC : 0);
}

/*
 * Hint to the kernel that the node no longer exists.
 * It won't immediately be killed from within the cache if a process
 * is still using it though. The kernel may even try to write back dirty
 * data.
 */
int
node_invalidnode(node)
	struct node *node;
{
	xfs_handle h;

	if (debugflag) {
		fprintf(stderr, "invalidnode(");
		node_repr(node, stderr);
		fprintf(stderr, ")\n");
	}
	node_to_handle(node, &h);
	return (*node->ctx->x_kops->invalidnode)(node->ctx, &h);
}

/*
 * Tell the kernel that the handle for a node has been changed.
 */
int
node_updatefid(old, new)
	struct node *old, *new;
{
	xfs_handle oh, nh;

	if (debugflag) {
		fprintf(stderr, "updatefid(old=");
		node_repr(old, stderr);
		fprintf(stderr, ", new=");
		node_repr(new, stderr);
		fprintf(stderr, ")\n");
	}

	node_to_handle(old, &oh);
	node_to_handle(new, &nh);
	return (*old->ctx->x_kops->updatefid)(old->ctx, &oh, &nh);
}
