/* $Id: nodeops.c,v 1.2 2000/02/29 04:15:50 leonard Exp $ */
/* Copyright David Leonard, 2000. All rights reserved. */
#include <stdio.h>
#include <stdlib.h>
#include <err.h>
#include <errno.h>
#include <string.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <dirent.h>
#include <xfs/xfs_message.h>
#include "xfs.h"
#include "debug.h"
#include "node.h"
#include "nodeops.h"

/*
 * Operations layer between XFS (above) and nodes below
 */

struct node *node_root;

/* Return 0 if 'cred' can access node 'n' for any access rights in 'right' */
static int
can_access(cred, n, right)
	xfs_cred *cred;
	struct node *n;
	int right;
{

#if 0
	mode_t m = n->stat.st_mode;
	/* 
	 * XXX this code is wrong. Should be using 'pag' structures and
	 * so forth.
	 */
	if (debugflag)
		fprintf(stderr, "access check mode=%03o owner=%d right=%s%s%s uid=%d\n",
			m, 
			n->stat.st_uid,
			(right & XFS_RIGHT_R) ? "r" : "",
			(right & XFS_RIGHT_W) ? "w" : "",
			(right & XFS_RIGHT_X) ? "x" : "",
			cred->uid
		);

	if ((right & XFS_RIGHT_R) && (m & S_IROTH)) return 0;
	if ((right & XFS_RIGHT_W) && (m & S_IWOTH)) return 0;
	if ((right & XFS_RIGHT_X) && (m & S_IXOTH)) return 0;

	if (cred->uid == 0 || (cred->uid == n->stat.st_uid)) {
		if ((right & XFS_RIGHT_R) && (m & S_IRUSR)) return 0;
		if ((right & XFS_RIGHT_W) && (m & S_IWUSR)) return 0;
		if ((right & XFS_RIGHT_X) && (m & S_IXUSR)) return 0;
	}

	errno = EACCES;
	return -1;
#else
	return 0;
#endif
}

static int
nodeop_version(octx, arg)
	struct xfs_opctx *octx;
	void *arg;
{
	return ENODEV;
}

static void
xa_to_stat(st, attr)
	struct stat *st;
	struct xfs_attr *attr;
{
	if (XA_VALID_MODE(attr))
		st->st_mode = attr->xa_mode;
	if (XA_VALID_NLINK(attr))
		st->st_nlink = attr->xa_nlink;
	if (XA_VALID_SIZE(attr))
		st->st_size = attr->xa_size;
	if (XA_VALID_UID(attr))
		st->st_uid = attr->xa_uid;
	if (XA_VALID_GID(attr))
		st->st_gid = attr->xa_gid;
	if (XA_VALID_ATIME(attr))
		st->st_atime = attr->xa_atime;
	if (XA_VALID_MTIME(attr))
		st->st_mtime = attr->xa_mtime;
	if (XA_VALID_CTIME(attr))
		st->st_ctime = attr->xa_ctime;
	if (XA_VALID_FILEID(attr))
		st->st_ino = attr->xa_fileid;
	if (XA_VALID_TYPE(attr)) {
		st->st_mode &= ~S_IFMT;
		switch (attr->xa_type) {
		case XFS_FILE_NON: break;
		case XFS_FILE_REG: st->st_mode |= S_IFREG; break;
		case XFS_FILE_DIR: st->st_mode |= S_IFDIR; break;
		case XFS_FILE_BLK: st->st_mode |= S_IFBLK; break;
		case XFS_FILE_CHR: st->st_mode |= S_IFCHR; break;
		case XFS_FILE_LNK: st->st_mode |= S_IFLNK; break;
		case XFS_FILE_SOCK: st->st_mode |= S_IFSOCK; break;
		case XFS_FILE_FIFO: st->st_mode |= S_IFIFO; break;
		case XFS_FILE_BAD: break;
		}
	}
}

/*
 * Request the root node with the given credentials.
 * This is usually the first thing that is called.
 * It must call xfs_installroot() before returning success.
 */
static int
nodeop_getroot(octx, cred)
	struct xfs_opctx *octx;
	xfs_cred *cred;
{
	int error;
	struct xfs_ctx *ctx;

	dprintf("-> GETROOT cred=%s\n", c2a(cred));

	ctx = xfs_opctx_to_ctx(octx);
	if (ctx != root_node->ctx)
		errx(1, "root node has wrong context");

	error = node_installroot(root_node);
	if (error)
		warnx("could not install root node: %s", strerror(error));

	return error;
}

/* 
 * Get the node named 'name' from the directory 'parent_handle'
 * Should call xfs_installnode() exactly once.
 */
static int
nodeop_getnode(octx, cred, parent_handle, name)
	struct xfs_opctx *octx;
	xfs_cred *cred;
	xfs_handle *parent_handle;
	const char *name;
{
	struct node *pn, *n;
	struct ncred ncred;

	dprintf("-> GETNODE cred=%s phnd=%s name=%s\n", c2a(cred), 
	    h2a(parent_handle), name);

	if ((pn = handle_to_node(parent_handle)) == NULL)
		return errno;

	if (!S_ISDIR(pn->stat.st_mode))
		return ENOTDIR;

	if (can_access(cred, pn, XFS_RIGHT_X))
		return errno;

	ncred.uid = cred->uid;

	if (strcmp(name, ".") == 0)
		n = pn;
	else if ((n = node_lookup(pn, name, &ncred)) == NULL)
		return errno;

	return node_installnode(pn, n, name);
}
	
/*
 * Get attributes for the node described by 'handle' (that was
 * previously installed) using credentials 'cred'
 */
static int
nodeop_getattr(octx, cred, handle)
	struct xfs_opctx *octx;
	xfs_cred *cred;
	xfs_handle *handle;
{
	struct node *n;

	dprintf("-> GETATTR cred=%s hnd=%s\n", c2a(cred), h2a(handle));

	if ((n = handle_to_node(handle)) == NULL)
		return errno;

	/* Note: anyone can access a file's attributes */

	if (node_getattr(n) == -1)
		return errno;
	n->tokens |= XFS_ATTR_R | XFS_ATTR_W;

	return node_installattr(n);
}

/*
 * Much like getattr, excepts gets the data for the node.
 * The data should be made available through a filehandle to a real file.
 * The 3rd arg to xfs_installdata should be the actual name for the filehandle,
 * but it is not used if the fh is valid.
 */
static int
nodeop_getdata(octx, cred, handle, tokens)
	struct xfs_opctx *octx;
	xfs_cred *cred;
	xfs_handle *handle;
	u_int tokens;
{
	struct node *n;

	dprintf("-> GETDATA cred=%s hnd=%s tok=%s\n", c2a(cred), h2a(handle),
	    t2a(tokens));

	if ((n = handle_to_node(handle)) == NULL)
		return errno;

	if (can_access(cred, n, XFS_RIGHT_R))	/* XXX FIXME */
		return errno;

	if (!n->file_valid) 
		if (node_alloc_file(n) == -1)
			return errno;
	if (node_getdata(n) == -1)
		return errno;

	n->stat.st_size = node_getsize_file(n);
	n->tokens |= tokens | XFS_DATA_R | XFS_DATA_W;

	return node_installdata(n, 0);
}

/*
 * Notification that the handle is no longer in the cache.
 * This function gets called some time after we notify the kernel
 * (with invalidnode) that we wanted to drop a node. ie this 
 * is a response to let us know if we can or can't remove the file.
 *
 * if flag & XFS_NOREFS then the node is still in the cache but unused.
 *   ie nobody wants the file but the kernel itself.
 * if flag & XFS_DELETE means the node has been dropped from the cache.
 *   ie the kernel has forgotten about the node.
 */
static void
nodeop_inactivenode(octx, handle, flag)
	struct xfs_opctx *octx;
	xfs_handle *handle;
	u_int flag;
{
	struct node *n;

	dprintf("-> INACTIVENODE hnd=%s\n", h2a(handle));
	if ((n = handle_to_node(handle)) == NULL)
		return;
	if (n->file_valid)
		node_dealloc_file(n);
	if (flag & XFS_DELETE)
		node_dealloc(n);
}

/*
 * Notification that a handle has been opened (with cred) to do what 
 * is described in tokens.
 * Can be the same as getdata() if no locking is implemented.
 */
static int
nodeop_open(octx, cred, handle, tokens)
	struct xfs_opctx *octx;
	xfs_cred *cred;
	xfs_handle *handle;
	u_int tokens;
{

	dprintf("-> OPEN cred=%s hnd=%s tok=%s\n", 
	    c2a(cred), h2a(handle), t2a(tokens));
	return nodeop_getdata(octx, cred, handle, tokens);
}

/*
 * This function is called on an fsync.
 * The data in the fh should then be shipped back to wherever
 * it came from.
 */
static int
nodeop_putdata(octx, handle, attr, cred, flag)
	struct xfs_opctx *octx;
	xfs_handle *handle;
	struct xfs_attr *attr;
	xfs_cred *cred;
	u_int flag;
{
	struct node *n;

	dprintf("-> PUTDATA hnd=%s attr=%s cred=%s flag=0x%u\n", 
	    h2a(handle), a2a(attr), c2a(cred), flag);

	if ((n = handle_to_node(handle)) == NULL)
		return errno;

	if (!n->file_valid) {
		warnx("nodeop_putdata: no backing file");
		return EIO;
	}

	/* Truncate the file to the new size */
	if (XA_VALID_SIZE(attr))
		n->stat.st_size = attr->xa_size;
	node_setsize_file(n, n->stat.st_size);

	if (node_putdata(n) == -1)
		return errno;
	return 0;
}


/* 
 * Called when attributes on a node are modified.
 * Note especially that this gets called when the size attribute of the
 * file changes.
 */
static int
nodeop_putattr(octx, handle, attr, cred)
	struct xfs_opctx *octx;
	xfs_handle *handle;
	struct xfs_attr *attr;
	xfs_cred *cred;
{
	struct node *n;
	struct stat newst;

	dprintf("-> PUTATTR hnd=%s attr=%s cred=%s\n",
	    h2a(handle), a2a(attr), c2a(cred));

	if ((n = handle_to_node(handle)) == NULL)
		return errno;
	if (can_access(cred, n, XFS_RIGHT_W))
		return errno;
	memcpy(&newst, &n->stat, sizeof newst);
	xa_to_stat(&newst, attr);
	if (node_putattr(n, &newst) == -1)
		return errno;
	return 0;
}

/* 
 * Called to create a new non-directory node named 'name' inside the directory 
 * 'parent_handle' using credentials 'cred' and default attrs 'attr'.
 */
static int
nodeop_create(octx, parent_handle, name, attr, mode, cred)
	struct xfs_opctx *octx;
	xfs_handle *parent_handle;
	const char *name;
	struct xfs_attr *attr;
	int mode;		/* unused */
	xfs_cred *cred;
{
	struct node *pn;
	struct stat newst;

	dprintf("-> CREATE phnd=%s name=%s attr=%s mode=%0o cred=%s\n",
	    h2a(parent_handle), name, a2a(attr), mode, c2a(cred));

	if ((pn = handle_to_node(parent_handle)) == NULL)
		return errno;
	if (!S_ISDIR(pn->stat.st_mode))
		return ENOTDIR;
	if (can_access(cred, pn, XFS_RIGHT_W))
		return errno;
	if (can_access(cred, pn, XFS_RIGHT_X))
		return errno;
	memcpy(&newst, &pn->stat, sizeof newst);
	xa_to_stat(&newst, attr);
	newst.st_mode = (newst.st_mode & ~ S_IFMT) | S_IFREG;
	if (node_create(pn, name, &newst) == -1)
		return errno;
	return 0;
}

/* 
 * Called to create a new subdirectory named 'name' inside the directory 
 * 'parent_handle' using credentials 'cred' and default attrs 'attr'.
 */
static int
nodeop_mkdir(octx, parent_handle, name, attr, cred)
	struct xfs_opctx *octx;
	xfs_handle *parent_handle;
	const char *name;
	struct xfs_attr *attr;
	xfs_cred *cred;
{
	struct node *pn;
	struct stat newst;

	dprintf("-> MKDIR phnd=%s name=%s attr=%s cred=%s\n",
	    h2a(parent_handle), name, a2a(attr), c2a(cred));

	if ((pn = handle_to_node(parent_handle)) == NULL)
		return errno;
	if (!S_ISDIR(pn->stat.st_mode))
		return ENOTDIR;
	if (can_access(cred, pn, XFS_RIGHT_W | XFS_RIGHT_X))
		return errno;
	memcpy(&newst, &pn->stat, sizeof newst);
	xa_to_stat(&newst, attr);
	newst.st_mode = (newst.st_mode & ~ S_IFMT) | S_IFDIR;
	if (node_create(pn, name, &newst) == -1)
		return errno;
	return 0;
}

/*
 * Create an alias to 'from_handle' inside directory 'parent_handle'
 * named 'name' using credentials 'cred'.
 */
static int
nodeop_link(octx, parent_handle, name, from_handle, cred)
	struct xfs_opctx *octx;
	xfs_handle *parent_handle;
	const char *name;
	xfs_handle *from_handle;
	xfs_cred *cred;
{

	dprintf("-> LINK phnd=%s name=%s\n", h2a(parent_handle), name);
	dprintf(" fhnd=%s cred=%s\n", h2a(from_handle), c2a(cred));
	return ENODEV;
}

/*
 * Create a symbolic link to 'contents'. The symlink should be named
 * 'name' inside directory 'parent_handle', and should have the given
 * attributes. All this done with credentials 'cred'.
 */
static int
nodeop_symlink(octx, parent_handle, name, contents, attr, cred)
	struct xfs_opctx *octx;
	xfs_handle *parent_handle;
	const char *name;
	const char *contents;
	struct xfs_attr *attr;
	xfs_cred *cred;
{

	dprintf("-> SYMLINK phnd=%s name=%s contents=%s attr=%s cred=%s\n",
	    h2a(parent_handle), name, contents, a2a(attr), c2a(cred));
	return ENOSYS;
}

/*
 * Remove the non-directory entry named 'name' from the 
 * directory 'parent_handle' using credentials 'cred'.
 */
static int
nodeop_remove(octx, parent_handle, name, cred)
	struct xfs_opctx *octx;
	xfs_handle *parent_handle;
	const char *name;
	xfs_cred *cred;
{
	struct node *pn, *n;
	struct ncred ncred;

	dprintf("-> REMOVE phnd=%s name=%s cred=%s\n",
	    h2a(parent_handle), name, c2a(cred));

	if (strcmp(name, ".") == 0 || strcmp(name, "..") == 0)
		return EPERM;

	if ((pn = handle_to_node(parent_handle)) == NULL)
		return errno;

	if (!S_ISDIR(pn->stat.st_mode))
		return ENOTDIR;
	if (can_access(cred, pn, XFS_RIGHT_X | XFS_RIGHT_W))
		return errno;

	ncred.uid = cred->uid;
	if ((n = node_lookup(pn, name, &ncred)) == NULL)
		return errno;
	if (S_ISDIR(n->stat.st_mode))
		return EISDIR;

	if (node_destroy(pn, name) == -1)
		return errno;

	return node_invalidnode(n);
}

/*
 * Remove the subdirectory entry named 'name' from the parent directory
 * 'parent_handle' using credentials 'cred'.
 * It is okay to return ENOTEMPTY.
 */
static int
nodeop_rmdir(octx, parent_handle, name, cred)
	struct xfs_opctx *octx;
	xfs_handle *parent_handle;
	const char *name;
	xfs_cred *cred;
{
	struct node *pn, *n;
	struct ncred ncred;

	dprintf("-> RMDIR phnd=%s name=%s cred=%s",
	    h2a(parent_handle), name, c2a(cred));

	if (strcmp(name, ".") == 0 || strcmp(name, "..") == 0)
		return EPERM;

	if ((pn = handle_to_node(parent_handle)) == NULL)
		return errno;

	if (!S_ISDIR(pn->stat.st_mode))
		return ENOTDIR;

	ncred.uid = cred->uid;
	if (can_access(cred, pn, XFS_RIGHT_X | XFS_RIGHT_W))
		return errno;

	if ((n = node_lookup(pn, name, &ncred)) == NULL)
		return errno;
	if (!S_ISDIR(n->stat.st_mode))
		return ENOTDIR;

	if (node_destroy(pn, name) == -1)
		return errno;

	return node_invalidnode(n);
}

/* 
 * Unlink the entry named 'old_name' from 'old_parent_handle', and re-link
 * it to the name 'new_name' inside 'new_parent_handle'. all done with
 * credentials 'cred'.
 */
static int
nodeop_rename(octx, old_parent_handle, old_name, new_parent_handle,
    new_name, cred)
	struct xfs_opctx *octx;
	xfs_handle *old_parent_handle;
	const char *old_name;
	xfs_handle *new_parent_handle;
	const char *new_name;
	xfs_cred *cred;
{
	struct node *opn, *npn, *n;
	struct ncred ncred;

	dprintf("-> RENAME ophnd=%s oname=%s", h2a(old_parent_handle), old_name);
	dprintf(" nphnd=%s nname=%s cred=%s\n", h2a(new_parent_handle), 
	    new_name, c2a(cred));

	if ((opn = handle_to_node(old_parent_handle)) == NULL)
		return errno;
	if ((npn = handle_to_node(new_parent_handle)) == NULL)
		return errno;

	if (!S_ISDIR(opn->stat.st_mode))
		return ENOTDIR;
	if (can_access(cred, opn, XFS_RIGHT_X | XFS_RIGHT_W))
		return errno;

	if (!S_ISDIR(npn->stat.st_mode))
		return ENOTDIR;
	if (can_access(cred, npn, XFS_RIGHT_X | XFS_RIGHT_W))
		return errno;

	ncred.uid = cred->uid;
	if ((n = node_lookup(opn, old_name, &ncred)) == NULL)
		return errno;

	/* Can return EXDEV if don't understand target npn */
	if (node_rename(opn, old_name, npn, new_name) == -1)
		return errno;

	return 0;
}

/*
 * Handle ioctl command 'opcode' on the file 'handle', using credentials 
 * 'cred'. The command input data is in 'inmsg' of length 'insize'.
 * There is space for the result data of length 'outsize' in the buffer
 * pointed to by 'outmsg'.
 */
static int
nodeop_pioctl(octx, opcode, cred, insize, outsize, inmsg, outmsg, handle)
	struct xfs_opctx *octx;
	int opcode;
	xfs_cred *cred;
	int insize;
	int outsize;
	const char *inmsg;
	char *outmsg;
	xfs_handle *handle;
{
	dprintf("-> PIOCTL op=0x%x cred=%s in[%d] out[%d] hndl=%s\n",
		opcode, c2a(cred), insize, outsize, h2a(handle));
	return ENOTTY;
}

/*
 * Handle an advisory lock on a file.
 * (Not implemented)
 */
static int
nodeop_advlock(octx, handle, cred, locktype, lockid)
	struct xfs_opctx *octx;
	xfs_handle *handle;
	xfs_cred *cred;
	xfs_locktype_t locktype;
	xfs_lockid_t lockid;
{
	dprintf("-> ADVLOCK hndl=%s cred=%s lktyp=%ld lkid=%ld\n",
		h2a(handle), c2a(cred), locktype, lockid);
	return ENOSYS;
}

struct xfs_user_ops nodeops = {
	nodeop_version,		/* version */
	nodeop_getroot,		/* getroot */
	nodeop_getnode,		/* getnode */
	nodeop_getattr,		/* getattr */
	nodeop_getdata,		/* getdata */
	nodeop_inactivenode,	/* inactivenode */
	nodeop_open,		/* open */
	nodeop_putdata,		/* putdata */
	nodeop_putattr,		/* putattr */
	nodeop_create,		/* create */
	nodeop_mkdir,		/* mkdir */
	nodeop_link,		/* link */
	nodeop_symlink,		/* symlink */
	nodeop_remove,		/* remove */
	nodeop_rmdir,		/* rmdir */
	nodeop_rename,		/* rename */
	nodeop_pioctl,		/* pioctl */
	nodeop_advlock		/* advlock */
};
