/* $Id: parse_ls.c,v 1.3 2000/02/29 04:15:50 leonard Exp $ */
/* Copyright David Leonard, 2000. All rights reserved. */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <time.h>
#include <sys/types.h>
#include <sys/stat.h>

extern int debugflag;

/*
 * Parse an 'ls -li' line like
 * "100159 -rw-r--r--  1 d  d     280 Feb 25 15:14 foo/bar/Makefile"
 * This routine tries to fill in the stat structure as best it 
 * can. The mtime/atime/ctime are set to the one time (mtime).
 * Also, the time is based on the current time, so the precision is
 * not very accurate. The current time is used to fill in unknowns.
 * Assumes the FTP server is in the same timezone as we are
 * The uid/gid fields are always set to zero.
 * The name pointer can be NULL if the name is unimportant.
 * Returns -1 if something didn't look right.
 */
int
parse_ls(line, linelen, st, namep, uidp, gidp, linkp)
	char *line;
	size_t linelen;
	struct stat *st;
	char **namep, **uidp, **gidp, **linkp;
{
	char *l, *link;
	int month;
	struct tm mtime_tm;
	static char *monthname[12] = {
		"Jan", "Feb", "Mar", "Apr", "May", "Jun",
		"Jul", "Aug", "Sep", "Oct", "Nov", "Dec" 
	};

	bzero(st, sizeof *st);
	l = line;

	while (*l == ' ') l++;
	if (isdigit(*l)) {
		st->st_ino = strtoul(l, &l, 10);	/* inode number */
		while (*l == ' ') l++;
	} else
		st->st_ino = 0;

	if      (l[0] == 'd') st->st_mode = S_IFDIR;	/* file type */
	else if (l[0] == 'b') st->st_mode = S_IFBLK;
	else if (l[0] == 'c') st->st_mode = S_IFCHR;
	else if (l[0] == 'l') st->st_mode = S_IFLNK;
	else if (l[0] == 'p') st->st_mode = S_IFIFO;
	else if (l[0] == 's') st->st_mode = S_IFSOCK;
	else if (l[0] == '-') st->st_mode = S_IFREG;
	else                  return -1;
	if      (l[1] == 'r') st->st_mode |= S_IRUSR;	/* permissions */
	else if (l[1] != '-') return -1;
	if      (l[2] == 'w') st->st_mode |= S_IWUSR;
	else if (l[2] != '-') return -1;
	if      (l[3] == 'x') st->st_mode |= S_IXUSR;
	else if (l[3] == 's') st->st_mode |= S_IXUSR | S_ISUID;
	else if (l[3] == 'S') st->st_mode |=           S_ISUID;
	else if (l[3] != '-') return -1;
	if      (l[4] == 'r') st->st_mode |= S_IRGRP;
	else if (l[4] != '-') return -1;
	if      (l[5] == 'w') st->st_mode |= S_IWGRP;
	else if (l[5] != '-') return -1;
	if      (l[6] == 'x') st->st_mode |= S_IXGRP;
	else if (l[6] == 's') st->st_mode |= S_IXGRP | S_ISGID;
	else if (l[6] == 'S') st->st_mode |=           S_ISGID;
	else if (l[6] != '-') return -1;
	if      (l[7] == 'r') st->st_mode |= S_IROTH;
	else if (l[7] != '-') return -1;
	if      (l[8] == 'w') st->st_mode |= S_IWOTH;
	else if (l[8] != '-') return -1;
	if      (l[9] == 'x') st->st_mode |= S_IXOTH;
	else if (l[9] == 't') st->st_mode |= S_IXOTH | S_ISTXT;
	else if (l[9] == 'T') st->st_mode |=           S_ISTXT;
	else if (l[9] != '-') return -1;
	l += 10;

	while (*l == ' ') l++;
	st->st_nlink = strtoul(l, &l, 10);		/* nlinks */

	while (*l == ' ') l++;
	if (uidp) *uidp = l;
	while (*l && *l != ' ') l++;			/* skip uid */
	if (*l) *l++ = '\0';
	st->st_uid = 0;

	while (*l == ' ') l++;
	if (gidp) *gidp = l;
	while (*l && *l != ' ') l++;			/* skip gid */
	if (*l) *l++ = '\0';
	st->st_gid = 0;

	while (*l == ' ') l++;
	st->st_size = strtoul(l, &l, 10);		/* file size */

	/*                                   01234567890123         */
	/* 100159 -rw-r--r--  1 d  d     280 Feb 25 15:14 Makefile  */
	/* 100159 -rw-r--r--  1 d  d     280 Feb 25  2000 Makefile  */

	/* Initialise with todays date */
	st->st_mtime = time(0);
	memcpy(&mtime_tm, localtime(&st->st_mtime), sizeof mtime_tm);

	while (*l == ' ') l++;
	for (month = 0; month < 12; month++)
		if (memcmp(monthname[month], l, 3) == 0)
			break;
	if (month < 12)  mtime_tm.tm_mon = month;
	else		 return -1;
	if (l[4] == ' ') mtime_tm.tm_mday = atoi(l+5);
	else             mtime_tm.tm_mday = atoi(l+4);
	if (l[9] == ':') {
		mtime_tm.tm_hour = atoi(l+7);
		mtime_tm.tm_min = atoi(l+10);
	} else
		mtime_tm.tm_year = atoi(l+8);
	l = l + 13;

	/* XXX seems no easy way to separate these times? */
	st->st_atime = st->st_ctime = st->st_mtime = mktime(&mtime_tm);

	if (namep || linkp) {
		/* trim off excessive trailing white space */
		while (line + linelen > l && isspace(line[linelen - 1]))
			linelen--;
		line[linelen] = '\0';

		if (S_ISLNK(st->st_mode)) {
			if ((link = strstr(l, " -> ")) != NULL) {
				*link = '\0';
				link += 4;
			}
		} else
			link = NULL;

		if (linkp)
			*linkp = link;
		if (namep)
			*namep = l;
	}

	if (debugflag)
		fprintf(stderr, "parse_ls: ino=%u mod=%o nlnk=%u own=%s:%s "
				"sz=%qu name=%s link=%s\n",
		    st->st_ino,
		    st->st_mode,
		    st->st_nlink,
		    uidp ? *uidp : "?",
		    gidp ? *gidp : "?",
		    st->st_size,
		    namep ? *namep : "?",
		    linkp ? (*linkp ? *linkp : "") : "?"
		);

	return 0;
}
