/* $Id: user.c,v 1.2 2000/02/29 04:15:50 leonard Exp $ */
/* Copyright David Leonard, 2000. All rights reserved. */
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <errno.h>
#include <fcntl.h>
#include <string.h>
#include <dirent.h>
#include <sys/stat.h>
#include <xfs/xfs_message.h>
#include "inum.h"
#include "xfs.h"
#include "dir.h"
#include "hash.h"
#include "user.h"
#include "null.h"
#include "ftproot.h"
#include "node.h"

/*
 * user: a directory full of usernames on a particular host
 */

struct user_node {
	struct node		node;
	const char 		*hostname;
	struct node		*parent;
	struct hash		*hash;
	int			passive;
};

static int
user_getdata(un)
	struct user_node *un;
{
	int fd;
	struct dirent de;
	char name[MAXNAMLEN];
	struct node *n;

	fd = node_open_file(&un->node, O_TRUNC|O_WRONLY);

	DIRENT_SET(&de, &un->node, ".");
	dir_write(fd, &de);

	DIRENT_SET(&de, un->parent, "..");
	dir_write(fd, &de);

	hash_first(un->hash);
	while (hash_next(un->hash, name, sizeof name, &n) == 0) {
		DIRENT_SET(&de, n, name);
		dir_write(fd, &de);
	}

	dir_close(fd);

	return 0;
}

static int
user_create(un, name, st)
	struct user_node *un;
	const char *name;
	struct stat *st;
{

	if (strcmp(name, ".") == 0)
		errno = EEXIST;
	else if (strcmp(name, "..") == 0)
		errno = EEXIST;
	else
		errno = EPERM;
	return -1;
}

static int
user_destroy(un, name)
	struct user_node *un;
	const char *name;
{
	struct node *n;

	if (strcmp(name, ".") == 0)
		errno = EPERM;
	else if (strcmp(name, "..") == 0)
		errno = EPERM;
	else if (hash_get(un->hash, name, &n) != 0)
		errno = ENOENT;
	else if (n->class != &ftproot_class)
		errno = EPERM;
	else if (ftproot_isconnected(n))
		errno = EBUSY;
	else if (hash_get(un->hash, name, &n) == 0) {
		if (hash_del(un->hash, name) == -1)
			return -1;
		node_invalidnode(n);
		un->node.tokens &= ~(XFS_DATA_R | XFS_DATA_W);
		return 0;
	}
	return -1;
}

static struct node *
user_lookup(un, name, cred)
	struct user_node *un;
	const char *name;
	struct ncred *cred;
{
	struct node *n;
	struct stat st;

	if (strcmp(name, ".") == 0)
		return &un->node;
	if (strcmp(name, "..") == 0)
		return un->parent;

	if (hash_get(un->hash, name, &n) == 0)
		return n;

	/* Don't allow any other dotfiles */
	if (name[0] == '.') {
		errno = ENOENT;
		return NULL;
	}

	st.st_uid = cred->uid;
	st.st_gid = 0;
	st.st_mode = S_IRWXU | S_IFDIR;

	/* Create user nodes on demand */
	if ((n = ftproot_new(&un->node, un->hostname, name, &st, un->passive)) 
	    == NULL)
		return NULL;
	if (hash_add(un->hash, name, n) == -1) {
		node_dealloc(n);
		return NULL;
	}
	un->node.tokens &= ~(XFS_DATA_R | XFS_DATA_W);
	node_installattr(&un->node);
	return n;
}

static void
user_dealloc(un)
	struct user_node *un;
{
	char name[MAXNAMLEN];
	struct node *n;

	if (un->hash) {
		hash_first(un->hash);
		while (hash_next(un->hash, name, sizeof name, &n) == 0)
			if (n != &un->node && n != un->parent)
				node_invalidnode(n);
		hash_free(un->hash);
		un->hash = NULL;
	}
	inum_reclaim(un->node.stat.st_ino);
	node_deinit(&un->node);
	free(un);
}

static void
user_repr(un, f)
	struct user_node *un;
	FILE *f;
{
	fprintf(stderr, "<user host=%s>", un->hostname);
}

struct node_class user_class = {
	(getattr_t)enoop,	/* getattr */
	(putattr_t)enoop,	/* putattr */
	user_getdata,		/* getdata */
	(putdata_t)enoop,	/* putdata */
	user_create,		/* create  */
	user_destroy,		/* destroy */
	user_lookup,		/* lookup  */
	(rename_t)enodev,	/* rename  */
	user_dealloc,		/* dealloc */
	user_repr		/* repr    */
};


struct node *
user_new(parent, hostname, passive)
	struct node *parent;
	const char *hostname;
	int passive;
{
	struct user_node *un;

	un = (struct user_node *)malloc(sizeof (struct user_node));
	if (un == NULL)
		return NULL;
	node_init(&un->node, &user_class, parent->ctx);
	un->hostname = hostname;
	if (parent->class == &null_class)
		parent = &un->node;
	un->parent = parent;
	un->passive = passive;
	un->hash = hash_new();
	un->node.stat.st_uid = 0;
	un->node.stat.st_gid = 0;
	un->node.stat.st_mode = S_IFDIR | S_IRWXU | S_IRWXG | S_IRWXO;
	un->node.stat.st_ino = inum_new();
	return (struct node *)un;
}
