/* $Id: xfs.c,v 1.1.1.1 2000/02/28 19:59:51 leonard Exp $ */
/* Copyright David Leonard, 2000. All rights reserved. */
#include <stdio.h>
#include <err.h>
#include <fcntl.h>
#include <stdlib.h>
#include <unistd.h>
#include <errno.h>
#include <pthread.h>
#include <pthread_np.h>
#include <string.h>
#include <sys/types.h>
#include <sys/queue.h>
#include <sys/syscall.h>
#include <sys/ioctl.h>
#include <xfs/xfs_message.h>
#include <xfs/xfs_debug.h>
#include <xfs/xfs_pioctl.h>
#include "xfs.h"
#include "debug.h"

#define _PATH_XFS "/dev/xfs0"

/* XXX only for debugging */
static char *xfs_msg_name[XFS_MSG_COUNT] = {
        "version", "wakeup", "getroot", "installroot", "getnode",
        "installnode", "getattr", "installattr", "getdata",
        "installdata", "inactivenode", "invalidnode", "open",
        "putdata", "putattr", "create", "mkdir", "link", "symlink",
        "remove", "rmdir", "rename", "pioctl", "wakeup_data",
        "updatefid", "advlock"
};
#define XFSMSGNAME(i) ((i) < 0 || (i) >= XFS_MSG_COUNT ? "?" : xfs_msg_name[i])

struct xfs_opctx {
	struct xfs_ctx	*o_ctx;
	struct xfs_message_header *o_hdr;
};

struct xfs_ctx *
xfs_opctx_to_ctx(octx)
	struct xfs_opctx *octx;
{
	return octx->o_ctx;
}

/* Send a one-way xfs message to the kernel. error code comes back in errno */
static int
xfs_send(ctx, msg)
	struct xfs_ctx *ctx;
	struct xfs_message_header *msg;
{
	static int seqno;
	static pthread_mutex_t mutex = PTHREAD_MUTEX_INITIALIZER;
	ssize_t len;
	int err;
	int error;

	if ((err = pthread_mutex_lock(&mutex)) != 0)
		errx(1, "pthread_mutex_lock: %s", strerror(err));

	msg->sequence_num = seqno++;

	len = write(ctx->x_fd, msg, msg->size);
	error = errno;

	if ((err = pthread_mutex_unlock(&mutex)) != 0)
		errx(1, "pthread_mutex_unlock: %s", strerror(err));

	if (len != -1 && len != msg->size)
		warnx("xfs: short write!");

	/* Return the error from xfs, if any */
	if (len == -1 && error) {
		dprintf("%s: error from xfs: %d %s\n", XFSMSGNAME(msg->opcode), 
		    error, strerror(error));
		return error;
	}
	return 0;
}

/* Send a WAKEUP reply message to the kernel */
void
xfs_reply(octx, error)
	struct xfs_opctx *octx;
	int error;
{
	struct xfs_message_wakeup w;

	w.header.opcode = XFS_MSG_WAKEUP;
	w.header.size = sizeof w;
	w.sleepers_sequence_num = octx->o_hdr->sequence_num;
	w.error = error;
	dprintf("<- WAKEUP sleeper=%d error=%d %s\n", 
	    w.sleepers_sequence_num, w.error, 
	    (w.error > 0 ? strerror(w.error) : ""));
	xfs_send(octx->o_ctx, &w.header);
}

/* Send a WAKEUP reply message to the kernel */
void
xfs_reply_data(octx, error, len, msg)
	struct xfs_opctx *octx;
	int error;
	int len;
	char *msg;
{
	struct xfs_message_wakeup_data w;

	w.header.opcode = XFS_MSG_WAKEUP_DATA;
	w.header.size = sizeof w - (sizeof w.msg - len);
	w.sleepers_sequence_num = octx->o_hdr->sequence_num;
	w.error = error;
	w.len = len;
	memcpy(w.msg, msg, len);
	dprintf("<- WAKEUP sleeper=%d len=%d error=%d %s\n", 
	    w.sleepers_sequence_num, w.len, w.error, 
	    (w.error > 0 ? strerror(w.error) : ""));
	xfs_send(octx->o_ctx, &w.header);
}

/* Send the INSTALLROOT message to the kernel */
static int
kop_installroot(ctx, node)
	struct xfs_ctx *ctx;
	struct xfs_msg_node *node;
{
	struct xfs_message_installroot m;

	m.header.opcode = XFS_MSG_INSTALLROOT;
	m.header.size = sizeof m;
	m.node = *node;
	return xfs_send(ctx, &m.header);
}

/* Send the INSTALLNODE message to the kernel */
static int
kop_installnode(ctx, parent_handle, name, node)
	struct xfs_ctx *ctx;
	xfs_handle *parent_handle;
	char *name;
	struct xfs_msg_node *node;
{
	struct xfs_message_installnode m;

	m.header.opcode = XFS_MSG_INSTALLNODE;
	m.header.size = sizeof m;
	m.parent_handle = *parent_handle;
	strlcpy(m.name, name, sizeof m.name);
	m.node = *node;
	return xfs_send(ctx, &m.header);
}

/* Send the INSTALLATTR message to the kernel */
static int
kop_installattr(ctx, node)
	struct xfs_ctx *ctx;
	struct xfs_msg_node *node;
{
	struct xfs_message_installattr m;

	m.header.opcode = XFS_MSG_INSTALLATTR;
	m.header.size = sizeof m;
	m.node = *node;
	return xfs_send(ctx, &m.header);
}

/*
 * Send the INSTALLDATA message to the kernel.
 * The cache_handle is something you can get from xfs_fhget().
 * Cache_handles to files are smaller than pathnames to carry around.
 */
static int
kop_installdata(ctx, node, cache_name, cache_handle, flag)
	struct xfs_ctx *ctx;
	struct xfs_msg_node *node;
	const char *cache_name;
	struct xfs_cache_handle *cache_handle;
	u_int flag;
{
	struct xfs_message_installdata m;

	m.header.opcode = XFS_MSG_INSTALLDATA;
	m.header.size = sizeof m;
	m.node = *node;
	strlcpy(m.cache_name, cache_name, sizeof m.cache_name);
	m.cache_handle = *cache_handle;
	m.flag = flag;
	return xfs_send(ctx, &m.header);
}

/* Send the INVALIDNODE message to the kernel */
static int
kop_invalidnode(ctx, handle)
	struct xfs_ctx *ctx;
	xfs_handle *handle;
{
	struct xfs_message_invalidnode m;

	m.header.opcode = XFS_MSG_INVALIDNODE;
	m.header.size = sizeof m;
	m.handle = *handle;
	return xfs_send(ctx, &m.header);
}

/* Send the UPDATEFID message to the kernel */
static int
kop_updatefid(ctx, old_handle, new_handle)
	struct xfs_ctx *ctx;
	xfs_handle *old_handle;
	xfs_handle *new_handle;
{
	struct xfs_message_updatefid m;

	m.header.opcode = XFS_MSG_UPDATEFID;
	m.header.size = sizeof m;
	m.old_handle = *old_handle;
	m.new_handle = *new_handle;
	return xfs_send(ctx, &m.header);
}

/* Kernel operations table */
static struct xfs_kernel_ops kops = {
	kop_installroot,
	kop_installnode,
	kop_installattr,
	kop_installdata,
	kop_invalidnode,
	kop_updatefid
};

static void *
receive(arg)
	void *arg;
{
	struct xfs_opctx *octx = (struct xfs_opctx *)arg; 
	struct xfs_ctx *ctx = octx->o_ctx;
	struct xfs_user_ops *u = ctx->x_uops;
	struct xfs_message_header *msg = octx->o_hdr;
	int error;
	int msglen = 0;
	char msgbuf[2048];

#if 0
	dprintf("receive op=%d\n", msg->opcode);
#endif
	error = ENODEV;

	pthread_set_name_np(pthread_self(), XFSMSGNAME(msg->opcode));

	switch (msg->opcode) {
	case XFS_MSG_VERSION:
		if (u->version)
			error = (*u->version)(octx, NULL);
		break;
	case XFS_MSG_GETROOT:
		if (u->getroot)
			error = (*u->getroot)(octx, 
			    &((struct xfs_message_getroot *)msg)->cred
			);
		break;
	case XFS_MSG_GETNODE:
		if (u->getnode)
			error = (*u->getnode)(octx, 
			    &((struct xfs_message_getnode *)msg)->cred,
			    &((struct xfs_message_getnode *)msg)->parent_handle,
			    ((struct xfs_message_getnode *)msg)->name
			);
		break;
	case XFS_MSG_GETATTR:
		if (u->getattr)
			error = (*u->getattr)(octx, 
			    &((struct xfs_message_getattr *)msg)->cred,
			    &((struct xfs_message_getattr *)msg)->handle
			);
		break;
	case XFS_MSG_GETDATA:
		if (u->getdata)
			error = (*u->getdata)(octx, 
			    &((struct xfs_message_getdata *)msg)->cred,
			    &((struct xfs_message_getdata *)msg)->handle,
			    ((struct xfs_message_getdata *)msg)->tokens
			);
		break;
	case XFS_MSG_INACTIVENODE:
		if (u->inactivenode)
			(*u->inactivenode)(octx, 
			    &((struct xfs_message_inactivenode *)msg)->handle,
			    ((struct xfs_message_inactivenode *)msg)->flag
			);
		break;
	case XFS_MSG_OPEN:
		if (u->open)
			error = (*u->open)(octx, 
			    &((struct xfs_message_open *)msg)->cred,
			    &((struct xfs_message_open *)msg)->handle,
			    ((struct xfs_message_open *)msg)->tokens
			);
		break;
	case XFS_MSG_PUTDATA:
		if (u->putdata)
			error = (*u->putdata)(octx, 
			    &((struct xfs_message_putdata *)msg)->handle,
			    &((struct xfs_message_putdata *)msg)->attr,
			    &((struct xfs_message_putdata *)msg)->cred,
			    ((struct xfs_message_putdata *)msg)->flag
			);
		break;
	case XFS_MSG_PUTATTR:
		if (u->putattr)
			error = (*u->putattr)(octx, 
			    &((struct xfs_message_putattr *)msg)->handle,
			    &((struct xfs_message_putattr *)msg)->attr,
			    &((struct xfs_message_putattr *)msg)->cred
			);
		break;
	case XFS_MSG_CREATE:
		if (u->create)
			error = (*u->create)(octx, 
			    &((struct xfs_message_create *)msg)->parent_handle,
			    ((struct xfs_message_create *)msg)->name,
			    &((struct xfs_message_create *)msg)->attr,
			    ((struct xfs_message_create *)msg)->mode,
			    &((struct xfs_message_create *)msg)->cred
			);
		break;
	case XFS_MSG_MKDIR:
		if (u->mkdir)
			error = (*u->mkdir)(octx, 
			    &((struct xfs_message_mkdir *)msg)->parent_handle,
			    ((struct xfs_message_mkdir *)msg)->name,
			    &((struct xfs_message_mkdir *)msg)->attr,
			    &((struct xfs_message_mkdir *)msg)->cred
			);
		break;
	case XFS_MSG_LINK:
		if (u->link)
			error = (*u->link)(octx, 
			    &((struct xfs_message_link *)msg)->parent_handle,
			    ((struct xfs_message_link *)msg)->name,
			    &((struct xfs_message_link *)msg)->from_handle,
			    &((struct xfs_message_link *)msg)->cred
			);
		break;
	case XFS_MSG_SYMLINK:
		if (u->symlink)
			error = (*u->symlink)(octx, 
			    &((struct xfs_message_symlink *)msg)->parent_handle,
			    ((struct xfs_message_symlink *)msg)->name,
			    ((struct xfs_message_symlink *)msg)->contents,
			    &((struct xfs_message_symlink *)msg)->attr,
			    &((struct xfs_message_symlink *)msg)->cred
			);
		break;
	case XFS_MSG_REMOVE:
		if (u->remove)
			error = (*u->remove)(octx, 
			    &((struct xfs_message_remove *)msg)->parent_handle,
			    ((struct xfs_message_remove *)msg)->name,
			    &((struct xfs_message_remove *)msg)->cred
			);
		break;
	case XFS_MSG_RMDIR:
		if (u->rmdir)
			error = (*u->rmdir)(octx, 
			    &((struct xfs_message_remove *)msg)->parent_handle,
			    ((struct xfs_message_remove *)msg)->name,
			    &((struct xfs_message_remove *)msg)->cred
			);
		break;
	case XFS_MSG_RENAME:
		if (u->rename)
			error = (*u->rename)(octx, 
			    &((struct xfs_message_rename *)msg)->
				old_parent_handle,
			    ((struct xfs_message_rename *)msg)->old_name,
			    &((struct xfs_message_rename *)msg)->
				new_parent_handle,
			    ((struct xfs_message_rename *)msg)->new_name,
			    &((struct xfs_message_rename *)msg)->cred
			);
		break;
	case XFS_MSG_PIOCTL:
		if (u->pioctl) {
			int insz = ((struct xfs_message_pioctl *)msg)->insize;
			msglen = ((struct xfs_message_pioctl *)msg)->outsize;
			if (insz > 2048) insz = 2048;
			if (msglen > 2048) msglen = 2048;
			error = (*u->pioctl)(octx,
			    ((struct xfs_message_pioctl *)msg)->opcode,
			    &((struct xfs_message_pioctl *)msg)->cred,
			    insz, msglen,
			    ((struct xfs_message_pioctl *)msg)->msg,
			    msgbuf,
			    &((struct xfs_message_pioctl *)msg)->handle
			);
		} else
			error = ENOTTY;
		break;
	case XFS_MSG_ADVLOCK:
		if (u->advlock)
			error = (*u->advlock)(octx, 
			    &((struct xfs_message_advlock *)msg)->handle,
			    &((struct xfs_message_advlock *)msg)->cred,
			    ((struct xfs_message_advlock *)msg)->locktype,
			    ((struct xfs_message_advlock *)msg)->lockid
			);
		break;
	default:
		warnx("unknown message %d received from xfs", msg->opcode);
		goto out;
	}

	if (msg->opcode == XFS_MSG_PIOCTL)
		xfs_reply_data(octx, error, msglen, msgbuf);
	if (msg->opcode != XFS_MSG_INACTIVENODE)
		xfs_reply(octx, error);

    out:
	free(octx);
	free(msg);
	return NULL;
}

struct xfs_ctx *
xfs_new(uops)
	struct xfs_user_ops *uops;
{
	struct xfs_ctx *ctx;
	int fd;

	fd = open(_PATH_XFS, O_RDWR);
	if (fd < 0)
		err(1, "open: %s", _PATH_XFS);

	if (debugflag)
		kdebug(fd, 
#if 0
/*			XDEBDEV         | */   /* device handling */
			XDEBMSG         |      /* message sending */
			XDEBDNLC        |      /* name cache */
			XDEBNODE        |      /* xfs nodes */
			XDEBVNOPS       |      /* vnode operations */
/*			XDEBVFOPS       | */   /* vfs operations */
			XDEBSYS         |      /* syscalls */
#endif
		0);
	else
		kdebug(fd, 0);

	ctx = (struct xfs_ctx *)malloc(sizeof (struct xfs_ctx));

	ctx->x_kops = &kops;
	ctx->x_uops = uops;
	ctx->x_fd = fd;

	dprintf("opened %s\n", _PATH_XFS);

	return ctx;
}

void
xfs_mainloop(ctx)
	struct xfs_ctx *ctx;
{
	struct xfs_opctx *octx;
	int error;
	char buffer[MAX_XMSG_SIZE];
	ssize_t len;
	char *b;
	struct xfs_message_header *h, *h2;
	pthread_t thid;
	pthread_attr_t attr;

	pthread_attr_init(&attr);
	pthread_attr_setdetachstate(&attr, PTHREAD_CREATE_DETACHED);
	while (1) {

		len = read(ctx->x_fd, buffer, sizeof buffer);
#if 0
		dprintf("read -> %d\n", len);
#endif
		if (len == -1 && errno == EINTR) {
			warn("read: %s", _PATH_XFS);
			continue;
		}
		if (len < 0)
			err(1, "read: %s", _PATH_XFS);
		if (len == 0)
			errx(1, "read: %s: closed", _PATH_XFS);

		b = buffer;
		while (len > 0) {
			h = (struct xfs_message_header *)b;
			b += h->size;
			len -= h->size;
			if (len < 0)
				errx(1, "short read");
			h2 = malloc(h->size);
			if (h2 == NULL)
				err(1, "malloc");
			octx = (struct xfs_opctx *)malloc(sizeof *octx);
			if (octx == NULL)
				err(1, "malloc");

			memcpy(h2, h, h->size);
			octx->o_ctx = ctx;
			octx->o_hdr = h2;

			error = pthread_create(&thid, &attr,
				receive, octx);

			if (error) {
				errx(1, "pthread_create: %s", strerror(error));
				/* free(h2); free(octx); */
			}

#if 0
			dprintf("thread %p handling msg seq=%d op=%d\n",
				thid, h->sequence_num, h->opcode);
#endif
		}
	}
}

/* Convert a filename into a cache_handle */
int
xfs_fhget(name, followsymlinks, cache_handle)
	const char *name;
	int followsymlinks;
	struct xfs_cache_handle *cache_handle;
{
	int error;
	struct ViceIoctl vi;

	vi.in_size = 0;
	vi.in = NULL;
	vi.out_size = sizeof *cache_handle;
	vi.out = (caddr_t)cache_handle;

	error = syscall(SYS_xfspioctl,
		AFSCALL_PIOCTL,		/* operation */
		name,			/* a_pathP */
		VIOC_FHGET,		/* a_opcode */
		&vi,			/* a_paramsP */
		followsymlinks		/* a_followSymlinks */
	);

	return error;

}

/* Open a file named by a cache_handle onto a file descriptor */
int
xfs_fhopen(cache_handle, flags)
	const struct xfs_cache_handle *cache_handle;
	int flags;				/* FREAD|FWRITE, etc */
{
	int fdesc;
	struct ViceIoctl vi;

	vi.in_size = sizeof *cache_handle;
	vi.in = (caddr_t)cache_handle;
	vi.out_size = 0;
	vi.out = NULL;

	fdesc = syscall(SYS_xfspioctl,
		AFSCALL_PIOCTL,		/* operation */
		NULL,			/* a_pathP */
		VIOC_FHOPEN,		/* a_opcode */
		&vi,			/* a_paramsP */
		flags			/* a_followSymlinks */
	);

	return fdesc;
}
